#!/usr/bin/env python
# coding: utf-8

import os
import unittest
import tempfile

from amulog import config
from amulog import lt_common


class TestCRF(unittest.TestCase):

    _path_trainfile = None
    _path_model = None

    @classmethod
    def setUpClass(cls):
        cls.data_train = [["ssh N D",
                           "login N D",
                           "failure N V",
                           "from N D",
                           "192.168.100.1 N V"],
                          ["su N D",
                           "user N D",
                           "sat N V",
                           "enabled N D"]]
        cls.data_test = ["ssh", "auth", "failure",
                         "from", "192.168.100.1", "user", "sat"]

        fd, cls._path_trainfile = tempfile.mkstemp()
        with os.fdopen(fd, 'w') as f:
            l_buf = []
            for lines in cls.data_train:
                l_buf.append("\n".join(lines))
            f.write("\n\n".join(l_buf))

        fd, cls._path_model = tempfile.mkstemp()
        os.close(fd)

    @classmethod
    def tearDownClass(cls):
        os.remove(cls._path_trainfile)
        os.remove(cls._path_model)

    def test_tagging(self):
        from amulog.alg.crf import init_ltgen
        conf = config.open_config(verbose=False)
        conf['log_template']['lt_methods'] = "shiso"
        conf["log_template_crf"]["model_filename"] = self._path_model

        table = lt_common.TemplateTable()
        ltgen = init_ltgen(conf, table)
        ltgen.init_trainer()
        ltgen.train_from_file(self._path_trainfile)

        tmp_pline = {"words": self.data_test}
        tpl = ltgen.generate_tpl(tmp_pline)
        self.assertTrue("ssh" in tpl)
        self.assertTrue(lt_common.REPLACER in tpl)


if __name__ == "__main__":
    unittest.main()





#!/usr/bin/env python
# coding: utf-8

"""Notice: after you edit log templates,
it is NOT reccommended to add other log messages,
especially with new log templates.
On current implementation, lt_edit will destroy
the structures of log template manager."""

import sys

from amulog import lt_common
from amulog import manager


def _str_lt(ltid, ld):
    return "ltid {0} : {1}".format(ltid, str(ld.lt(ltid)))


def merge_ltid(ld, ltid1, ltid2):
    ltm = manager.LTManager(ld.conf, ld.db, ld.lttable)
    print("merge following log templates...")
    print(_str_lt(ltid1, ld))
    print(_str_lt(ltid2, ld))
    print()

    ltw1 = ld.lt(ltid1).ltw
    cnt1 = ld.lt(ltid1).count
    l_s = ld.lt(ltid1).lts
    ltw2 = ld.lt(ltid2).ltw
    cnt2 = ld.lt(ltid2).count
    if not len(ltw1) == len(ltw2):
        sys.exit("log template length is different, failed")
    new_ltw = lt_common.merged_template(ltw1, ltw2)

    ltm.replace_lt(ltid1, new_ltw, l_s, cnt1 + cnt2)
    ltm.remove_lt(ltid2)
    ld.db.update_log({"ltid": ltid2}, ltid=ltid1)

    ld.commit_db()

    print("> new log template : ltid {0}".format(ltid1))
    print(_str_lt(ltid1, ld))


def separate_ltid(ld, ltid, vid, value):

    def _remake_lt(_ltid):
        # use iter_words rather than iter_lines
        # because iter_lines needs ltline registered in table
        # but ltline on new_ltid is now on construction in this function...
        new_lt = None
        cnt = 0
        for l_w in ld.db.iter_words(ltid=_ltid):
            if new_lt is None:
                new_lt = l_w
            else:
                new_lt = lt_common.merged_template(new_lt, l_w)
            cnt += 1
        print("result : " + str(new_lt))
        print()
        return new_lt, cnt

    ltm = manager.LTManager(ld.conf, ld.db, ld.lttable)
    print("separate following log template...")
    print(_str_lt(ltid, ld))
    print("new log template if variable {0} is {1}".format(vid, value))
    print()

    l_lid = [lm.lid for lm in ld.iter_lines(ltid=ltid)
             if lm.var()[vid] == value]
    new_ltid = ld.lttable.next_ltid()
    for lid in l_lid:
        ld.db.update_log({"lid": lid}, ltid=new_ltid)

    l_s = ld.lt(ltid).lts
    ltw1, cnt1 = _remake_lt(ltid)
    ltm.replace_lt(ltid, ltw1, l_s, cnt1)

    ltw2, cnt2 = _remake_lt(new_ltid)
    ltline = ltm.add_lt(ltw2, l_s, cnt2)
    assert ltline.ltid == new_ltid

    ld.commit_db()

    print("> new log templates : ltid {0}, ltid {1}".format(ltid, new_ltid))
    print(_str_lt(ltid, ld))
    print(_str_lt(new_ltid, ld))


def split_ltid(ld, ltid, vid):
    ltm = manager.LTManager(ld.conf, ld.db, ld.lttable)
    print("split following log template...")
    print(_str_lt(ltid, ld))

    d_lid = {}
    for lm in ld.iter_lines(ltid=ltid):
        lid = lm.lid
        w = lm.var()[vid]
        d_lid.setdefault(w, []).append(lid)

    print("variable {0}: {1}".format(vid, d_lid.keys()))
    if len(d_lid) > 10:
        ret = input("split into {0} new templates, ok? [y/n]: ").lower()
        if ret not in ("y", "yes"):
            return

    args = list(d_lid.keys())
    d_word_ltid = {}

    ltobj = ld.lt(ltid)
    l_s = ltobj.lts
    vloc = ltobj.var_location()[vid]

    # Update a template on existing ltid
    new_ltid = ltid
    new_word = args.pop(0)
    new_ltw = ltobj.ltw[:]
    new_ltw[vloc] = new_word
    cnt = len(d_lid[new_word])
    ltm.replace_lt(new_ltid, new_ltw, l_s, cnt)
    d_word_ltid[new_word] = new_ltid
    print("> new log templates : ltid {0}".format(new_ltid))
    print(_str_lt(new_ltid, ld))

    # Add templates
    while len(args) > 0:
        new_ltid = ld.lttable.next_ltid()
        new_word = args.pop(0)
        new_ltw = ltobj.ltw[:]
        new_ltw[vloc] = new_word
        cnt = len(d_lid[new_word])
        new_ltobj = ltm.add_lt(new_ltw, l_s, cnt)
        assert new_ltid == new_ltobj.ltid
        d_word_ltid[new_word] = new_ltid
        print("> new log templates : ltid {0}".format(new_ltid))
        print(_str_lt(new_ltid, ld))

    # Update log lines
    for word, l_lid in d_lid.items():
        temp_ltid = d_word_ltid[word]
        for lid in l_lid:
            ld.db.update_log({"lid": lid}, ltid=temp_ltid)

    ld.commit_db()


def fix_ltid(ld, ltid, l_vid):
    ltm = manager.LTManager(ld.conf, ld.db, ld.lttable)
    print("make variable (with no variety) into description word...")
    print(_str_lt(ltid, ld))

    d_variety = {vid: set() for vid in l_vid}
    for lm in ld.iter_lines(ltid=ltid):
        for vid in l_vid:
            d_variety[vid].add(lm.var()[vid])

    d_fixed = {}
    for vid in l_vid:
        variety = d_variety[vid]
        assert len(variety) > 0
        if len(variety) == 1:
            print("confirmed that variable {0} is stable".format(vid))
            d_fixed[vid] = variety.pop()
            print("fixed word : {0}".format(d_fixed[vid]))
        else:
            print("variable {0} is not stable, ignored".format(vid))
            print(variety)

    ltobj = ld.lt(ltid)
    new_ltw = ltobj.ltw[:]
    for vid in d_fixed.keys():
        vloc = ltobj.var_location()[vid]
        new_ltw[vloc] = d_fixed[vid]
    l_s = ltobj.lts
    cnt = ltobj.count

    ltm.replace_lt(ltid, new_ltw, l_s, cnt)
    ld.commit_db()
    print("> new log templates : ltid {0}".format(ltid))
    print(_str_lt(ltid, ld))


def free_ltid(ld, ltid, l_wid):
    ltm = manager.LTManager(ld.conf, ld.db, ld.lttable)
    print("make description word into variable (with no variety)...")
    print(_str_lt(ltid, ld))

    ltobj = ld.lt(ltid)
    new_ltw = ltobj.ltw[:]
    for wid in l_wid:
        if ltobj.ltw[wid] == lt_common.REPLACER:
            print("wid {0} seems a variable, ignored")
        else:
            print("confirmed that wid {0} is description word ({1})".format(
                wid, ltobj.ltw[wid]))
            new_ltw[wid] = lt_common.REPLACER
    l_s = ltobj.lts
    cnt = ltobj.count

    ltm.replace_lt(ltid, new_ltw, l_s, cnt)
    ld.commit_db()
    print("> new log templates : ltid {0}".format(ltid))
    print(_str_lt(ltid, ld))



import os
import pickle
import sys
import logging
from typing import Optional
from importlib import import_module

import log2seq
from amulog import config
from amulog import strutil
from amulog import lt_common
from amulog import log_db
from amulog import host_alias

_logger = logging.getLogger(__package__)

ONLINE_COMMIT_INTERVAL = 1000
_MULTIPROCESS_LOCAL_OBJECTS = dict()


class LTManager(object):
    """A manager to generate log templates and their database.

    Attributes:
        #TODO

    """

    # adding lt to db (ltgen do not add)

    def __init__(self, conf, db, lttable, reset_db=False, parallel=False):
        self._conf = conf
        self._reset_db = reset_db
        self._filename = conf["manager"]["indata_filename"]
        self._fail_output = conf["manager"]["fail_output"]
        self._online_batchsize = conf.getint("manager", "online_batchsize")
        self._online_counter = 0
        self._offline_batchsize = conf.getint("manager", "offline_batchsize")
        self._drop_undefhost = conf.getboolean("manager", "undefined_host")
        self._shuffle_import = conf.getboolean("log_template_import", "shuffle")

        self._db = db
        self._lttable = lttable
        self._table = lt_common.TemplateTable()
        self._ltgen: Optional[lt_common.LTGen] = None

        self._pool = None
        if parallel:
            tmp_n_proc = conf["manager"]["n_process"]
            if tmp_n_proc.isdigit():
                n_proc = int(tmp_n_proc)
            else:
                n_proc = os.cpu_count()
            ltgen_kwargs = {"conf": conf,
                            "table": None,  # individual table for child process
                            "shuffle": self._shuffle_import}
            from multiprocessing import Pool
            self._pool = Pool(processes=n_proc,
                              initializer=self._init_pool,
                              initargs=(ltgen_kwargs,))
        else:
            self._lp = load_log2seq(self._conf)
            self._ha = host_alias.init_hostalias(self._conf)
            self._drop_undefhost = conf.getboolean("manager", "undefined_host")
            self._ltgen = init_ltgen_methods(self._conf, self._table)

        self._ltgroup = init_ltgroup(self._conf, self._lttable)
        if not self._reset_db:
            self._ltgroup.restore_ltg(self._db, self._lttable)

    @property
    def template_table(self):
        return self._table

    @staticmethod
    def _init_pool(ltgen_kwargs):
        objects = {}
        conf = ltgen_kwargs["conf"]
        objects["lp"] = load_log2seq(conf)
        objects["ha"] = host_alias.init_hostalias(conf)
        objects["drop_undefhost"] = conf.getboolean("manager", "undefined_host")

        ltgen = init_ltgen_methods(**ltgen_kwargs)
        assert not ltgen.is_stateful(), \
            "parallel processing is limited to stateless methods"
        objects["ltgen"] = ltgen

        global _MULTIPROCESS_LOCAL_OBJECTS
        _MULTIPROCESS_LOCAL_OBJECTS = objects

    @staticmethod
    def _pool_task(batch):
        lp = _MULTIPROCESS_LOCAL_OBJECTS["lp"]
        ha = _MULTIPROCESS_LOCAL_OBJECTS["ha"]
        ltgen = _MULTIPROCESS_LOCAL_OBJECTS["ltgen"]
        drop_undefhost = _MULTIPROCESS_LOCAL_OBJECTS["drop_undefhost"]

        ret = []
        for message_id, line in batch:
            pline = parse_line(strutil.add_esc(line), lp)
            pline = normalize_pline(pline, ha, drop_undefhost)
            if pline is None:
                ret.append([message_id, None, None])
            else:
                tpl = ltgen.generate_tpl(pline)
                ret.append([message_id, pline, tpl])
        return ret

    def _process_offline_parallel(self, iterable_lines):
        def _sigterm_handler():
            raise KeyboardInterrupt

        import signal
        signal.signal(signal.SIGTERM, _sigterm_handler)

        l_input = [(mid, line)
                   for mid, line in enumerate(iterable_lines)]
        iter_batch = (l_input[i::self._offline_batchsize]
                      for i in range(self._offline_batchsize))

        try:
            d_pline = {}
            d_tid = {}
            for ret in self._pool.imap_unordered(self._pool_task, iter_batch):
                for mid, pline, tpl in ret:
                    d_pline[mid] = pline
                    if tpl is None:
                        continue
                    if self._table.exists(tpl):
                        tid = self._table.get_tid(tpl)
                    else:
                        tid = self._table.add(tpl)
                    d_tid[mid] = tid
            self._pool.close()
        except KeyboardInterrupt:
            self._pool.terminate()
            exit()
        else:
            return d_pline, d_tid

    def _process_offline_single(self, iterable_lines):
        d_pline = {}
        for mid, line in enumerate(iterable_lines):
            pline = parse_line(strutil.add_esc(line), self._lp)
            pline = normalize_pline(pline, self._ha, self._drop_undefhost)
            d_pline[mid] = pline

        offline_input = {mid: pline for mid, pline in d_pline.items()
                         if pline is not None}
        d_tid = self._ltgen.process_offline(offline_input)
        return d_pline, d_tid

    def process_offline(self, list_lines):
        if self._pool is None:
            d_pline, d_tid = self._process_offline_single(list_lines)
        else:
            if (not self._reset_db) and self._ltgen.is_stateful():
                msg = ("offline additional change is limited "
                       "to stateless ltgen methods")
                raise ValueError(msg)
            d_pline, d_tid = self._process_offline_parallel(list_lines)

        s_added = set()
        for mid, pline in d_pline.items():
            if pline is None or mid not in d_tid:
                self.fail_dump(list_lines[mid])
                continue

            tid = d_tid[mid]
            if tid is None:
                self.fail_dump(list_lines[mid])
                continue
            elif tid in s_added:
                ltid = self._table.get_ltid(tid)
                self.count_lt(ltid)
                ltline = self._lttable[ltid]
            else:
                tpl = self._table.get_template(tid)
                ltline = self.add_lt(tpl, pline[log2seq.KEY_SYMBOLS])
                self._table.add_ltid(tid, ltline.ltid)
                s_added.add(tid)
            self.add_line(pline, ltline)

        self.commit_db()
        self.dump()

    def get_parsed_line(self, line):
        pline = parse_line(strutil.add_esc(line), self._lp)
        pline = normalize_pline(pline, self._ha, self._drop_undefhost)
        return pline

    def process_line(self, line):
        pline = self.get_parsed_line(line)
        if pline is None:
            self.fail_dump(line)
            return None

        tid, state = self._ltgen.process_line(pline)
        if tid is None:
            self.fail_dump(line)
            return None
        elif state == lt_common.LTGen.state_added:
            tpl = self._ltgen.get_tpl(tid)
            ltline = self.add_lt(tpl, pline[log2seq.KEY_SYMBOLS],
                                 add_group=True)
            self._table.add_ltid(tid, ltline.ltid)
        elif state == lt_common.LTGen.state_changed:
            tpl = self._ltgen.get_tpl(tid)
            ltid = self._table.get_ltid(tid)
            self.replace_and_count_lt(ltid, tpl)
            ltline = self._lttable[ltid]
        elif state == lt_common.LTGen.state_unchanged:
            ltid = self._table.get_ltid(tid)
            self.count_lt(ltid)
            ltline = self._lttable[ltid]
        else:
            raise AssertionError

        self.add_line(pline, ltline)
        self._online_counter += 1
        if self._online_counter >= self._online_batchsize:
            self.commit_db()
            self._online_counter = 0
        return ltline

    def process_online_end(self):
        if isinstance(self._ltgroup, lt_common.LTGroupOffline):
            self.remake_ltg()

    def add_line(self, pline, ltline):
        """Add a log message to DB.

        Args:
            ltline (lt_common.LogTemplate): A log template object.
            pline (dict): A parsed log message with log2seq.

        Returns:
            LogMessage: An annotated log message instance.
        """
        dt = pline[log2seq.KEY_TIMESTAMP]
        host = pline["host"]
        l_w = pline[log2seq.KEY_WORDS]
        kwargs = {"ltid": ltline.ltid,
                  "dt": dt,
                  "host": host,
                  "l_w": l_w}
        if "lid" in pline:
            kwargs["lid"] = pline["lid"]
        new_lid = self._db.add_line(**kwargs)
        return log_db.LogMessage(new_lid, ltline,
                                 dt, host, l_w)

    def add_lt(self, l_w, l_s, cnt=1, add_group=False):
        # add new lt to db and table
        ltid = self._lttable.next_ltid()
        ltline = lt_common.LogTemplate(ltid, None, l_w, l_s, cnt)
        # if self._ltgroup is not None:
        if add_group and isinstance(self._ltgroup, lt_common.LTGroupOnline):
            ltgid = self._ltgroup.add(ltline)
        else:
            ltgid = ltid
        ltline.ltgid = ltgid
        self._lttable.add_lt(ltline)
        self._db.add_lt(ltline)
        self._db.add_ltg(ltline.ltid, ltline.ltgid)
        return ltline

    def replace_lt(self, ltid, l_w, l_s=None, cnt=None):
        self._lttable[ltid].replace(l_w, l_s, cnt)
        self._db.update_lt(ltid, l_w, l_s, cnt)

    def replace_and_count_lt(self, ltid, l_w, l_s=None):
        cnt = self._lttable[ltid].increment()
        self._lttable[ltid].replace(l_w, l_s, None)
        self._db.update_lt(ltid, l_w, l_s, cnt)

    def count_lt(self, ltid):
        cnt = self._lttable[ltid].increment()
        self._db.update_lt(ltid, None, None, cnt)

    def remove_lt(self, ltid):
        self._lttable.remove_lt(ltid)
        self._db.remove_lt(ltid)

    def remake_ltg(self):
        self._db.reset_ltg()
        self._lttable = self._ltgroup.make()
        for ltline in self._lttable:
            self._db.add_ltg(ltline.ltid, ltline.ltgid)

    def load_internal_data(self):
        if os.path.exists(self._filename) and not self._reset_db:
            self.load()

    def commit_db(self):
        """Commit requested changes in LogDB.
        """
        self._db.commit()

    def load(self):
        with open(self._filename, 'rb') as f:
            obj = pickle.load(f)
        table_data, ltgen_data, ltgroup_data = obj
        self._table.load(table_data)
        self._ltgen.load(ltgen_data)
        self._ltgroup.load(ltgroup_data)

    def dump(self):
        table_data = self._table.dumpobj()
        if self._ltgen is None:
            ltgen_data = None
        else:
            ltgen_data = self._ltgen.dumpobj()
        if self._ltgroup is None:
            ltgroup_data = None
        else:
            ltgroup_data = self._ltgroup.dumpobj()
        obj = (table_data, ltgen_data, ltgroup_data)
        with open(self._filename, 'wb') as f:
            pickle.dump(obj, f)

    def fail_dump(self, msg):
        with open(self._fail_output, 'a') as f:
            f.write(msg)


def init_ltgen_methods(conf, table, lt_methods=None, shuffle=None):
    if lt_methods is None:
        lt_methods = config.getlist(conf, "log_template", "lt_methods")
    if shuffle is None:
        shuffle = conf.getboolean("log_template_import", "shuffle")

    if len(lt_methods) > 1:
        l_ltgen = []
        import_index = None
        for index, method_name in enumerate(lt_methods):
            ltgen = init_ltgen(conf, table, method_name, shuffle)
            l_ltgen.append(ltgen)
            if method_name == "import":
                import_index = index
        return lt_common.LTGenJoint(table, l_ltgen, import_index)
    elif len(lt_methods) == 1:
        return init_ltgen(conf, table, lt_methods[0], shuffle)
    else:
        raise ValueError


def init_ltgen(conf, table, method, shuffle=False):
    kwargs = {"conf": conf,
              "table": table,
              "shuffle": shuffle}
    if method == "import":
        from . import lt_import
        return lt_import.init_ltgen_import(**kwargs)
    elif method == "import-ext":
        from . import lt_import_ext
        return lt_import_ext.init_ltgen_import_ext(**kwargs)
    elif method == "crf":
        from amulog.alg.crf import lt_crf
        return lt_crf.init_ltgen_crf(**kwargs)
    elif method == "re":
        from amulog import lt_regex
        return lt_regex.init_ltgen_regex(**kwargs)
    elif method == "va":
        from . import lt_va
        return lt_va.init_ltgen_va(**kwargs)
    else:
        modname = "amulog.alg." + method
        alg_module = import_module(modname)
        return alg_module.init_ltgen(**kwargs)


def init_ltgroup(conf, lttable):
    ltg_alg = conf.get("log_template", "ltgroup_alg")
    if ltg_alg == "shiso":
        from amulog.alg.shiso import shiso
        ltgroup = shiso.LTGroupSHISO(lttable,
                                     ngram_length=conf.getint(
                                            "log_template_shiso", "ltgroup_ngram_length"),
                                     th_lookup=conf.getfloat(
                                            "log_template_shiso", "ltgroup_th_lookup"),
                                     th_distance=conf.getfloat(
                                            "log_template_shiso", "ltgroup_th_distance"),
                                     mem_ngram=conf.getboolean(
                                            "log_template_shiso", "ltgroup_mem_ngram")
                                     )
    elif ltg_alg == "ssdeep":
        from . import lt_misc
        ltgroup = lt_misc.LTGroupFuzzyHash(lttable)
    elif ltg_alg == "semantics":
        from . import ltg_semantics
        ltgroup = ltg_semantics.init_ltgroup_semantics(conf, lttable)
    elif ltg_alg == "none":
        ltgroup = None
    else:
        raise ValueError("ltgroup_alg({0}) invalid".format(ltg_alg))

    return ltgroup


def load_log2seq(conf):
    """Return parser object log2seq.LogParser .

    Amulog accepts following additional keys extracted by log2seq.
    - host: device hostname, mandatory
    - lid: Log message identifier, optional

    The configuration is described in a python script file.
    The file is specified in database.parser_script option,
    and the "parser" module variable in the script is returned.
    Otherwise, this function returns log2seq default parser.

    Args:
        conf: config object.

    Returns:
        log2seq.LogParser

    """
    fp = conf.get("manager", "parser_script")
    if len(fp.strip()) == 0:
        return log2seq.init_parser()
    else:
        import sys
        from importlib import import_module
        path = os.path.dirname(fp)
        sys.path.append(os.path.abspath(path))
        libname = os.path.splitext(os.path.basename(fp))[0]
        script_mod = import_module(libname)
        lp = script_mod.parser
        return lp


def parse_line(line, lp):
    try:
        parsed_line = lp.process_line(line)
    except log2seq.LogParseFailure:
        return None
    else:
        if len(parsed_line[log2seq.KEY_WORDS]) == 0:
            msg = "pass empty message {0}".format(line.strip("\n"))
            _logger.debug(msg)
            return None
        else:
            return parsed_line


def normalize_pline(pline, ha, drop_undefhost=False):
    if pline is None:
        return None

    if "lid" in pline:
        pline["lid"] = int(pline["lid"])

    org_host = pline["host"]
    host = ha.resolve_host(org_host)
    if host is None:
        if drop_undefhost:
            return None
        else:
            host = org_host
    pline["host"] = host

    return pline


def _open_file(fp, **kwargs):
    ext = os.path.splitext(fp)[-1].lstrip(".")
    if ext == "bz2":
        import bz2
        open_func = bz2.open
    elif ext == "gz":
        import gzip
        open_func = gzip.open
    else:
        ext = "text"
        open_func = open

    _logger.info("processing {0} file {1}".format(ext, fp))
    return open_func(fp, 'rt', **kwargs)


def iter_lines(targets, encoding="utf-8", errors="ignore"):
    for fp in targets:
        if os.path.isdir(fp):
            sys.stderr.write(
                "{0} is a directory, fail to process\n".format(fp))
            sys.stderr.write(
                "Use -r if you need to search log data recursively\n")
        else:
            if not os.path.isfile(fp):
                raise IOError("File {0} not found".format(fp))
            with _open_file(fp, encoding=encoding, errors=errors) as f:
                for line in f:
                    yield line


def iter_plines(conf, targets, pass_none=True):
    lp = load_log2seq(conf)
    ha = host_alias.init_hostalias(conf)
    drop_undefhost = conf.getboolean("manager", "undefined_host")

    for line in iter_lines(targets):
        pline = parse_line(strutil.add_esc(line), lp)
        pline = normalize_pline(pline, ha, drop_undefhost)
        if pline is None and pass_none:
            pass
        else:
            yield pline


def process_files_online(conf, targets, reset_db):
    """Add log messages to DB from files.

    Args:
        conf (config.ExtendedConfigParser): A common configuration object.
        targets (List[str]): A sequence of filepaths to process.
        reset_db (bool): True if DB needs to reset before adding.

    Raises:
        IOError: If a file in targets not found.
    """
    def _sigterm_handler():
        raise KeyboardInterrupt

    import signal
    signal.signal(signal.SIGTERM, _sigterm_handler)

    msg = "amulog online processing"
    _logger.info(msg)

    ld = log_db.LogData(conf, edit=True, reset_db=reset_db)
    ltm = LTManager(conf, ld.db, ld.lttable, reset_db=reset_db)

    try:
        for line in iter_lines(targets):
            ltm.process_line(line)
    except KeyboardInterrupt:
        pass
    finally:
        ltm.process_online_end()
        ltm.commit_db()
        ltm.dump()


def process_files_offline(conf, targets, reset_db, parallel=False):
    """Add log messages to DB from files. This function do NOT process
    messages incrementally. Use this to avoid bad-start problem of
    log template generation with clustering or training methods.

    Note:
        This function needs large memory space.

    Args:
        conf (config.ExtendedConfigParser): A common configuration object.
        targets (List[str]): A sequence of filepaths to process.
        reset_db (bool): True if DB needs to reset before adding.
        parallel (bool, optional): Use multiprocessing.

    Raises:
        IOError: If a file in targets not found.
    """
    msg = "amulog offline processing"
    if parallel:
        msg += " in parallel"
    _logger.info(msg)

    ld = log_db.LogData(conf, edit=True, reset_db=reset_db)
    ltm = LTManager(conf, ld.db, ld.lttable, reset_db=reset_db,
                    parallel=parallel)

    l_line = [line for line in iter_lines(targets)]
    ltm.process_offline(l_line)


def data_from_data(conf, targets, dirname, method, reset):
    rod = log_db.RestoreOriginalData(dirname, method=method, reset=reset)
    lp = load_log2seq(conf)
    ha = host_alias.init_hostalias(conf)
    drop_undefhost = conf.getboolean("database", "undefined_host")
    for line in iter_lines(targets):
        pline = parse_line(line, lp)
        pline = normalize_pline(pline, ha, drop_undefhost)
        if pline is None:
            continue
        rod.add_str(pline[log2seq.KEY_TIMESTAMP], line)
    rod.commit()


def remake_ltgroup(conf):
    ld = log_db.LogData(conf, edit=True, reset_db=False)
    ltm = LTManager(conf, ld.db, ld.lttable, reset_db=False)
    ltm.remake_ltg()
    ltm.commit_db()



#!/usr/bin/env python
# coding: utf-8

import os
import logging

from . import lt_common
from .external import tpl_match
from .external import mod_tplseq
from .external import regexhash

_logger = logging.getLogger(__package__)


class LTGenImportExternal(lt_common.LTGenStateless):

    def __init__(self, table, filename, mode, mode_esc, ltmap, head, shuffle=False):
        super(LTGenImportExternal, self).__init__(table)
        self._table = table
        self._fp = filename
        self._l_tpl = self._load_tpl(self._fp, mode, mode_esc)
        self._l_regex = [tpl_match.generate_regex(tpl)
                         for tpl in self._l_tpl]
        if ltmap == "hash":
            self._rtable = regexhash.RegexHashTable(self._l_tpl, self._l_regex,
                                                    head)
        elif ltmap == "table":
            self._rtable = regexhash.RegexTable(self._l_tpl, self._l_regex)
        else:
            raise NotImplementedError

        if shuffle:
            self._rtable.shuffle()

    @staticmethod
    def _load_tpl(fp, mode, mode_esc):
        l_tpl = []
        if not os.path.exists(fp):
            errmes = ("log_template_import.def_path"
                      " {0} is invalid".format(fp))
            raise ValueError(errmes)
        with open(fp, 'r') as f:
            for line in f:
                if mode == "plain":
                    mes = line.rstrip("\n")
                elif mode == "ids":
                    line = line.rstrip("\n")
                    mes = line.partition(" ")[2].strip()
                else:
                    raise ValueError("invalid import_mode {0}".format(mode))
                if len(mes) == 0:
                    continue
                if mode_esc:
                    l_tpl.append(mes)
                else:
                    l_tpl.append(tpl_match.add_esc_external(mes))
        return l_tpl

    def generate_tpl(self, pline):
        mes = pline["message"]
        ret = self._rtable.search(mes)
        if ret is None:
            _logger.debug(
                "No log template found for message : {0}".format(mes))
            return None
        else:
            tplid, matchobj = ret
            tpl = self._l_tpl[tplid]
            new_tpl = mod_tplseq.redefine_tpl(tpl, pline, matchobj=matchobj)
            return new_tpl

    #def process_line(self, pline):
    #    mes = pline["message"]
    #    ret = self._rtable.search(mes)
    #    if ret is None:
    #        _logger.debug(
    #            "No log template found for message : {0}".format(mes))
    #        return None, None
    #    else:
    #        tplid, matchobj = ret
    #        tpl = self._rtable.l_tpl[tplid]
    #        new_tpl = mod_tplseq.redefine_tpl(tpl, pline, self.sym,
    #                                          matchobj=matchobj)

    #        if self._table.exists(new_tpl):
    #            tid = self._table.get_tid(new_tpl)
    #            return tid, self.state_unchanged
    #        else:
    #            tid = self._table.add(new_tpl)
    #            return tid, self.state_added


def init_ltgen_import_ext(conf, table, shuffle, **kwargs):
    fn = conf.get("log_template_import", "def_path_ext")
    mode = conf.get("log_template_import", "import_format_ext")
    if fn == "":
        fn = conf.get("log_template_import", "def_path")
    mode_esc = conf.getboolean("log_template_import", "import_format_ext_esc")
    ltmap = conf.get("log_template_import", "ext_search_method")
    head = conf.getint("log_template_import", "hash_strlen")

    return LTGenImportExternal(table, fn, mode, mode_esc, ltmap, head, shuffle)



#!/usr/bin/env python
# coding: utf-8


def load_trainitems(fp):
    """Load labeled train data and output items"""

    with open(fp, 'r') as f:
        buf = []
        for line in f:
            if line.strip() == "":
                if len(buf) > 0:
                    yield buf
                    buf = []
            else:
                item = line.strip().split()
                assert len(item) == 3
                buf.append(item)
        else:
            if len(buf) > 0:
                yield buf


def dump_trainitems(iterable_items, fp):
    with open(fp, 'w') as f:
        f.write("\n\n".join([items2str(lineitems)
                             for lineitems in iterable_items]))


def items2str(lineitems):
    return "\n".join([" ".join(item) for item in lineitems])


def line2items(l_w, midlabel_func=None, dummy_label="N"):
    """Returns items of a line with dummy label, use only for tagging"""
    if midlabel_func is None:
        midlabel_func = lambda x: x
    return [(w, midlabel_func(w), dummy_label) for w in l_w]

#
# def make_trainitems(l_w, tpl, midlabel_func=None):
#     """Returns items of a line for training from log_db.LogMessage()"""
#     if midlabel_func is None:
#         midlabel_func = lambda x: x
#     ret = []
#     for line_w, tpl_w in zip(l_w, tpl):
#         if line_w == tpl_w:
#             label = "D"
#         else:
#             label = "V"
#         ret.append((line_w, midlabel_func(line_w), label))
#     return ret
#
#
# def lm2trainitems(lm: log_db.LogMessage, midlabel_func=None):
#     return make_trainitems(lm.l_w, lm.lt.ltw, midlabel_func)



#!/usr/bin/env python
# coding: utf-8

import os
import time
import datetime
import logging
import subprocess  # for python3


# args
json_args = {"ensure_ascii": False,
             "indent": 4,
             "sort_keys": True}


# classes

#class singleton(object):
#
#    def __new__(clsObj, *args, **kwargs):
#        tmpInstance = None
#        if not hasattr(clsObj, "_instanceDict"):
#            clsObj._instanceDict = {}
#            clsObj._instanceDict[str(hash(clsObj))] = \
#                super(singleton, clsObj).__new__(clsObj, *args, **kwargs)
#            tmpInstance = clsObj._instanceDict[str(hash(clsObj))]
#        elif not hasattr(clsObj._instanceDict, str(hash(clsObj))):
#            clsObj._instanceDict[str(hash(clsObj))] = \
#                super(singleton, clsObj).__new__(clsObj, *args, **kwargs)
#            tmpInstance = clsObj._instanceDict[str(hash(clsObj))]
#        else:
#            tmpInstance = clsObj._instanceDict[str(hash(clsObj))]
#        return tmpInstance


class IDDict:

    def __init__(self, keyfunc=None):
        self._d_obj = {}
        self._d_id = {}
        if keyfunc is None:
            self.keyfunc = lambda x: x
        else:
            self.keyfunc = keyfunc

    def _next_id(self):
        next_id = len(self._d_obj)
        assert next_id not in self._d_obj
        return next_id

    def add(self, obj):
        if self.exists(obj):
            return self._d_id[self.keyfunc(obj)]
        else:
            key = self._next_id()
            self.set_item(key, obj)
            return key

    def set_item(self, key, obj):
        self._d_obj[key] = obj
        self._d_id[self.keyfunc(obj)] = key

    def exists(self, obj):
        return self.keyfunc(obj) in self._d_id

    def get(self, keyid):
        return self._d_obj[keyid]


# file managing

def is_empty(dirname):
    if os.path.isdir(dirname):
        if len(os.listdir(dirname)) > 1:
            return True
        else:
            return False
    else:
        return False


def rep_dir(args):
    if isinstance(args, list):
        ret = []
        for arg in args:
            if os.path.isdir(arg):
                ret.extend(["/".join((arg, fn)) \
                            for fn in sorted(os.listdir(arg))])
            else:
                ret.append(arg)
        return ret
    elif isinstance(args, str):
        arg = args
        if os.path.isdir(arg):
            return ["/".join((arg, fn)) for fn in sorted(os.listdir(arg))]
        else:
            return [arg]
    else:
        raise NotImplementedError


def recur_dir(args):
    def open_path(path):
        if os.path.isdir(path):
            ret = []
            for fn in sorted(os.listdir(path)):
                ret += open_path("/".join((path, fn)))
            return ret
        else:
            return [path]

    if isinstance(args, list):
        l_fn = []
        for arg in args:
            l_fn += open_path(arg)
    elif isinstance(args, str):
        l_fn = open_path(args)
    else:
        raise NotImplementedError
    return l_fn


def filepath(dn, fn):
    if len(dn) == 0:
        return fn
    else:
        return "/".join((dn, fn))


def filename(fp):
    if "/" in fp:
        return fp.split("/")[-1]
    else:
        return fp


def mkdir(path):
    if not os.path.exists(path):
        os.mkdir(path)
    elif not os.path.isdir(path):
        raise OSError("something already exists on given path, "
                      "and it is NOT a directory")
    else:
        pass


def rm(path):
    if os.path.exists(path):
        os.remove(path)
        return True
    else:
        return False


def rm_dirchild(dirpath):
    for fpath in rep_dir(dirpath):
        os.remove(fpath)


def filepath_local(current_path, fn):
    return os.path.dirname(os.path.abspath(current_path)) + "/" + fn


def last_modified(args, latest=False):
    """Get the last modified time of a file or a set of files.

    Args:
        args (str or list[str]): Files to investigate.
        latest (Optional[bool]): If true, return the latest datetime
                of timestamps. Otherwise, return the oldest timestamp.

    Returns:
        datetime.datetime

    """

    def file_timestamp(fn):
        stat = os.stat(fn)
        t = stat.st_mtime
        return datetime.datetime.fromtimestamp(t)

    if isinstance(args, list):
        l_dt = [file_timestamp(fn) for fn in args]
        if latest:
            return max(l_dt)
        else:
            return min(l_dt)
    elif isinstance(args, str):
        return file_timestamp(args)
    else:
        raise NotImplementedError


# subprocess
def call_process(cmd):
    """Call a subprocess and handle standard outputs.
    
    Args:
        cmd (list): A sequence of command strings.
    
    Returns:
        ret (int): Return code of the subprocess.
        stdout (str)
        stderr (str)
    """

    p = subprocess.Popen(cmd, shell=True,
                         stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = p.communicate()
    ret = p.returncode
    return ret, stdout, stderr


# parallel computing

def mprocess(l_process, pal):
    """
    Handle multiprocessing with an upper limit of working processes.
    This function incrementally fill working processes with remaining tasks.
    This function does not use memory sharing or process communications.
    If you need processe communications,
    use common.mprocess_queueing.

    Args:
        l_process (List(multiprocess.Process))
        pal (int): Maximum number of processes executed at once.
    """

    l_job = []
    while len(l_process) > 0:
        if len(l_job) < pal:
            job = l_process.pop(0)
            job.start()
            l_job.append(job)
        else:
            time.sleep(1)
            l_job = [j for j in l_job if j.is_alive()]
    else:
        for job in l_job:
            job.join()


def mprocess_queueing(l_pq, n_process):
    """
    Same as mprocess, but this function yields
    returned values of every processes with multiprocessing.Queue.

    Args:
        l_pq (List[multiprocessing.Process, multiprocessing.Queue])
        n_process (int): Maximum number of processes executed at once.
    """

    l_job = []
    while len(l_pq) > 0:
        if len(l_job) < n_process:
            process, queue = l_pq.pop(0)
            process.start()
            l_job.append((process, queue))
        else:
            time.sleep(1)
            l_temp = []
            for process, queue in l_job:
                if queue.empty():
                    l_temp.append((process, queue))
                else:
                    ret = queue.get()
                    yield ret
                    assert queue.empty()
                    process.join()
                    queue.close()
            l_job = l_temp
    else:
        while len(l_job) > 0:
            time.sleep(1)
            l_temp = []
            for process, queue in l_job:
                if queue.empty():
                    l_temp.append((process, queue))
                else:
                    ret = queue.get()
                    yield ret
                    assert queue.empty()
                    process.join()
                    queue.close()
            l_job = l_temp


# def mthread_queueing(l_thread, pal):
#    """
#    Args:
#        l_thread (List[threading.Thread]): A sequence of thread objects.
#        pal (int): Maximum number of threads executed at once.
#    """
#    l_job = []
#    while len(l_thread) > 0:
#        if len(l_job) < pal:
#            job = l_thread.pop(0)
#            job.start()
#            l_job.append(job)
#        else:
#            time.sleep(1)
#            l_job = [j for j in l_job if j.is_alive()]
#    else:
#        for job in l_job:
#            job.join()
#
#
# def mprocess_queueing(l_process, pal):
#    mthread_queueing(l_process, pal)


# measurement

class Timer:

    def __init__(self, header, output=None, timestr_func=None):
        self._dts = None
        self._dte = None
        self._lap_dt = []
        self.header = header
        self.output = output
        self._timestr_func = None

    def _output(self, string):
        if isinstance(self.output, logging.Logger):
            self.output.info(string)
        else:
            print(string)

    def timestr(self, td):
        if self._timestr_func is None:
            return td.total_seconds()
        else:
            return self._timestr_func(td)

    def start(self):
        self._dts = datetime.datetime.now()
        self._lap_dt.append(self._dts)
        self._output("{0} start".format(self.header))

    def lap(self, name):
        if self._dts is None:
            raise ValueError("call start() before lap()")
        lap_dt = datetime.datetime.now()
        t = self.timestr(lap_dt - self._dts)
        self._lap_dt.append(lap_dt)
        self._output("{0} lap({1}) ({2})".format(self.header, name, t))

    def lap_diff(self, name):
        if self._dts is None:
            raise ValueError("call start() before lap_diff()")
        lap_dt = datetime.datetime.now()
        t = self.timestr(lap_dt - self._lap_dt[-1])
        self._lap_dt.append(lap_dt)
        self._output("{0} diff({1}) ({2})".format(self.header, name, t))

    def stop(self):
        if self._dts is None:
            raise ValueError("call start() before stop()")
        self._dte = datetime.datetime.now()
        t = self.timestr(self._dte - self._dts)
        self._output("{0} done ({1})".format(self.header, t))

    def total_time(self):
        if self._dts is None or self._dte is None:
            raise ValueError("call start() and stop() before total_time()")
        else:
            return self._dte - self._dts

    def stat(self):
        if len(self._lap_dt) == 1:
            raise ValueError
        import numpy as np
        lap_times = np.diff(self._lap_dt)
        avg = np.average(lap_times)
        se = np.std(lap_times) / np.sqrt(len(lap_times))
        self._output("{0} lap times: {0}".format(self.header, lap_times))
        self._output("{0} average: {0}".format(self.header, avg))
        self._output("{0} standard error: {0}".format(self.header, se))


# visualization

def show_repr(iterable, head, foot, indent=0, strfunc=str):
    data = list(iterable)
    if (head <= 0 and foot <= 0) or \
            (len(data) <= head + foot):
        buf = [strfunc(e) for e in data]
    else:
        buf = []
        if head > 0:
            buf += [strfunc(e) for e in data[:head]]
        buf += [" " * indent + "..."]
        if foot > 0:
            buf += [strfunc(e) for e in data[-foot:]]
        buf.append("({0})".format(len(data)))
    header = " " * indent
    return "\n".join([header + line for line in buf])


def cli_table(table, spl=" ", fill=" ", align="left"):
    """
    Args:
        table (List[List[str]]): input data
        spl (str): splitting string of columns
        fill (str): string of 1 byte, used to fill the space
        align (str): left and right is available
    """
    len_column = len(table[0])
    col_max = [0] * len_column

    for row in table:
        for cid, val in enumerate(row):
            if cid >= len_column:
                raise ValueError
            val = str(val)
            if len(val) > col_max[cid]:
                col_max[cid] = len(val)

    l_buf = []
    for row in table:
        line = []
        for cid, val in enumerate(row):
            cell = str(val)
            len_cell = col_max[cid]
            len_space = len_cell - len(cell)
            if align == "left":
                cell = cell + fill * len_space
            elif align == "right":
                cell = fill * len_space + cell
            else:
                raise NotImplementedError
            line.append(cell)
        l_buf.append(spl.join(line))

    return "\n".join(l_buf)


# compatibility

def pickle_comp_args(comp):
    if comp:
        d = {"encoding": "bytes"}
    else:
        d = {}
    return d



#!/usr/bin/env python
# coding: utf-8

"""VA (Vaarandi's algorithm) is a simplified algorithm of SLCT and LogCluster.
It estimates log templates on the basis of an assumption
'description words appear more frequently than variable words'.
"""

from collections import defaultdict

from . import lt_common


class LTGenVA(lt_common.LTGen):
    """
    Notice:
        In this log template generation algorithm,
        incremental process will fail in the beginning
        part of input data set.
        To avoid this, it is strongly recommended to use
        process_init_data (i.e., db-make-init)
        instead of process_line (i.e., db-make)
        with enough size of initial data.
    """

    def __init__(self, table, method, threshold):
        super().__init__(table)
        self.method = method
        self.th = threshold
        self._d_wordcnt = defaultdict(int)

    def load(self, loadobj):
        self._d_wordcnt = loadobj

    def dumpobj(self):
        return self._d_wordcnt

    def _add_dict(self, l_w):
        self._d_wordcnt[None] += 1 # line count
        for w in l_w:
            self._d_wordcnt[w] += 1

    def _label(self, l_w):
        if self.method == "static":
            assert int(self.th) > 1, "lt_va.threshold is too small"
            l_label = []
            for w in l_w:
                if self._d_wordcnt[w] > self.th:
                    l_label.append(True)
                else:
                    l_label.append(False)
            return l_label
        elif self.method == "relative-line":
            assert self.th < 1, "lt_va.threshold is too large"
            l_label = []
            for w in l_w:
                if self._d_wordcnt[w] > self.th * self._d_wordcnt[None]:
                    l_label.append(True)
                else:
                    l_label.append(False)
            return l_label
        elif self.method == "relative-variable":
            # if th = 0.5, description words >= variable words
            assert self.th < 1, "lt_va.threshold is too large"
            l_cnt = [self._d_wordcnt[w] for w in l_w]
            temp_th = sorted(l_cnt)[int(self.th * len(l_w))]
            l_label = []
            for w in l_w:
                if self._d_wordcnt[w] >= temp_th:
                    l_label.append(True)
                else:
                    l_label.append(False)
            return l_label
        else:
            raise NotImplementedError

    def _label2tpl(self, l_w, l_label):
        ret = []
        for w, label in zip(l_w, l_label):
            if label is True:
                ret.append(w)
            elif label is False:
                ret.append(lt_common.REPLACER)
            else:
                raise ValueError
        return ret

    def preprocess(self, plines):
        for pline in plines:
            l_w = pline["words"]
            self._add_dict(l_w)

    def process_line(self, pline):
        l_w = pline["words"]
        self._add_dict(l_w)
        l_label = self._label(l_w)
        tpl = self._label2tpl(l_w, l_label)
        return self.update_table(tpl)


def init_ltgen_va(conf, table, **_):
    method = conf.get("log_template_va", "method")
    threshold = conf.getfloat("log_template_va", "threshold")
    return LTGenVA(table, method, threshold)





#!/usr/bin/env python
# coding: utf-8

"""
A log template generation algorithm proposed in [1].
[1] P. He, et al. Drain: An Online Log Parsing Approach with Fixed Depth Tree. ICWS 2017, pp.33–40, 2017.

After editing log templates manually (with lt_tool),
Drain do not work correctly.
Do NOT edit log templates manually if you still have unprocessed log data.
"""

import os
from amulog import lt_common

DEFAULT_REGEX_CONFIG = "/".join((os.path.dirname(os.path.abspath(__file__)),
                                 "../../data/drain_regex.conf"))


class Node:

    def __init__(self):
        self.child = {}
        self.clusters = None  # set of tid


class LTGenDrain(lt_common.LTGen):
    """
    Args:
        table (LogTemplateTable): see lt_common.LTGen
        threshold (int): Threshold value used in 2nd step (clustering).
        depth (int): Tree depth used in 1st part (preceding tokens).
            (depth - 2) top words except known variables
            are always considered NOT variable.
        vreobj (lt_regex.VariableRegex): Regex-based variable classifier.
            Mainly used in 1st step (preceding tokens).
    """

    def __init__(self, table, threshold, depth, vreobj,
                 count_replacer=False):
        super().__init__(table)
        self._threshold = threshold
        self._depth = depth
        self._vre = vreobj
        self._count_replacer = count_replacer
        self._root = Node()

    def load(self, loadobj):
        self._root = loadobj

    def dumpobj(self):
        return self._root

    def process_line(self, pline):
        # preprocess
        tokens = [lt_common.REPLACER if self._vre.match(w) else w
                  for w in pline["words"]]

        # search by length
        length = len(tokens)
        if length not in self._root.child:
            self._root.child[length] = Node()
        node = self._root.child[length]

        # search by preceding tokens
        ptokens = [w for w in tokens if w != lt_common.REPLACER]
        for i in range(self._depth - 2):
            try:
                key = ptokens[i]
            except IndexError:
                key = None
            if key not in node.child:
                node.child[key] = Node()
            node = node.child[key]

        # search by token similarlity
        max_sim_seq = 0
        tid = None
        if node.clusters is None:
            node.clusters = set()
        for tmp_tid in node.clusters:
            tpl = self._table[tmp_tid]
            sim_seq = sum(1 for t, w in zip(tokens, tpl) if t == w) / length
            if sim_seq >= self._threshold and sim_seq > max_sim_seq:
                max_sim_seq = sim_seq
                tid = tmp_tid

        # update clusters
        if tid is None:
            tid = self.add_tpl(tokens)
            node.clusters.add(tid)
            state = self.state_added
        else:
            state = self.merge_tpl(tokens, tid)

        return tid, state


def init_ltgen(conf, table, **_):
    threshold = conf.getfloat("log_template_drain", "threshold")
    depth = conf.getint("log_template_drain", "depth")

    from amulog.lt_regex import VariableRegex
    preprocess_fn = conf.get("log_template_drain", "preprocess_rule")
    if preprocess_fn.strip() == "":
        preprocess_fn = DEFAULT_REGEX_CONFIG
    vreobj = VariableRegex(conf, preprocess_fn)

    return LTGenDrain(table, threshold, depth, vreobj)


def get_param_candidates():
    import numpy as np
    from itertools import product
    params = []
    for th, depth in product(np.arange(0, 1.1, 0.1), (3, 4)):
        params.append({"threshold": th,
                       "depth": depth})
    return params


def init_ltgen_with_params(conf, table, params, **_):
    threshold = params["threshold"]
    depth = params["depth"]

    from amulog.lt_regex import VariableRegex
    preprocess_fn = conf.get("log_template_drain", "preprocess_rule")
    if preprocess_fn.strip() == "":
        preprocess_fn = DEFAULT_REGEX_CONFIG
    vreobj = VariableRegex(conf, preprocess_fn)

    return LTGenDrain(table, threshold, depth, vreobj)



#!/usr/bin/env python
# coding: utf-8


"""DEPRECATED"""

# Preprocessing for structuralizing log messages

import sys
import os
import re
import datetime
import ipaddress

from . import config
from . import strutil

DEFAULT_SYMDEF = "/".join((os.path.dirname(__file__), "data/symdef.txt.default"))


class LogParser():

    re_time = re.compile(r"^\d{2}:\d{2}:\d{2}(\.\d+)?$")
    re_mac = re.compile(r"^[0-9a-fA-F]{2}:[0-9a-fA-F]{2}:[0-9a-fA-F]{2}:[0-9a-fA-F]{2}:[0-9a-fA-F]{2}:[0-9a-fA-F]{2}$")
    #l_var_re = [re_time, re_mac, re_ipv6]
    l_var_re = [re_time, re_mac]

    re_datetime = re.compile(r"^\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}")
    re_year = re.compile(r"^[12]\d{3}$")
    month_name = ("Jan", "Feb", "Mar", "Apr", "May", "Jun",
            "Jul", "Aug", "Sep", "Oct", "Nov", "Dec")

    def __init__(self, conf, sep_variable = False):
        self.default_year = conf.get("database", "default_year")
        self.rmheader_fl = config.gettuple(conf, "database",
                                           "remove_header_filename")
        self.symdef = conf.get("log_template", "sym_filename")
        self.varsym = conf.get("log_template", "variable_symbol")
        if self.symdef is None or self.symdef == "":
            self.symdef = DEFAULT_SYMDEF
        self.sep_variable = sep_variable
        self.sym_ignore = conf.getboolean("log_template", "sym_ignore")
        hsize = conf.get("database", "header_size")
        if hsize == "" or hsize is None:
            self.header_size = None
        else:
            self.header_size = int(hsize)

        self.rm_header = []
        if self.rmheader_fl is not None and len(self.rmheader_fl) > 0:
            self._init_remove_header(self.rmheader_fl)
        self._init_splitter()

    def _init_remove_header(self, l_fn):
        for fn in l_fn:
            with open(fn, 'r'):
                for line in f:
                    line = line.rstrip("\n")
                    if not line == "":
                        self.rm_header.append(line.rstrip("\n"))

    def _init_splitter(self):
        buf = []
        with open(self.symdef, "r") as f:
            for line in f:
                buf.append(line.rstrip("\n").split("0"))
        self.spl = buf[0] # splitter
        self.cspl = buf[1] # conditional splitter (except variables in RE)

    def _is_removed(self, line):
        for description in self.rm_header:
            if description in line:
                return True
        else:
            return False

    def _set_year(self):
        if self.default_year is None or self.default_year == "":
            return datetime.datetime.today().year
        else:
            return int(self.default_year)
    
    def _re_cspl(self, string):
        # judge string is a variable or complex word to be parted
        try:
            r = ipaddress.ip_address(string)
        except ValueError:
            try:
                r = ipaddress.ip_network(string, strict = False)
            except ValueError:
                pass
            else:
                return True
        else:
            return True

        for restr in self.l_var_re:
            if restr.match(string):
                return True
        else:
            return False

    def _part(self, string):
        # partition if string have splitter
        for cnt, c in enumerate(string):
            if c in self.spl:
                return string[0:cnt], string[cnt], string[cnt+1:]
        # partition if string have variable string symbol
        # only for Non-symbol-ignoring mode
        if self.varsym in string and self.sep_variable:
            ind_s = string.find(self.varsym)
            ind_e = ind_s + len(self.varsym)
            if string == self.varsym:
                return string, None, None
            elif ind_s == 0:
                # top of string is equal to self.varsym
                return string[0:ind_e], "", string[ind_e:]
            else:
                return string[0:ind_s], "", string[ind_s:]
        # partition in reversed order if string is not one variable
        for cnt in reversed(range(len(string))):
            if string[cnt] in self.cspl:
                if self._re_cspl(string):
                    return string, None, None
                else:
                    return string[0:cnt], string[cnt], string[cnt+1:]
        return string, None, None

    def _split_word(self, string):
        # partition string recursively and label them
        w1, s, w2 = self._part(string)
        if w2 is None:
            assert s is None
            return [(w1, 'w')]
        else:
            ret1 = self._split_word(w1)
            ret2 = self._split_word(w2)
            return ret1 + [(s, 's')] + ret2

    @staticmethod
    def _merge_sym(l_elem):
        # merge continuous symbol strings
        # input : splited strings and labels(w or s)
        # output : sequence of words and symbol strings
        l_w = []
        l_s = []
        temp = []
        for e in l_elem:
            if e[1] == 'w':
                if len(e[0]) > 0:
                    l_s.append("".join(temp))
                    temp = []
                    l_w.append(e[0])
            elif e[1] == 's':
                temp.append(e[0])
            else:
                raise ValueError
        else:
            l_s.append("".join(temp))
        assert len(l_w) + 1 == len(l_s)
        return l_w, l_s

    def split_message(self, line):
        ret = self._split_word(line.rstrip("\n"))
        l_w, l_s = self._merge_sym(ret)
        if self.sym_ignore:
            return l_w, l_s
        else:
            # restore order of l_w and l_s, and return it as l_w (l_s is None)
            # s[0], w[0], s[1], w[1], ..., s[n], w[n], s[n+1]
            ret = []
            for w, s in zip(l_w + [""], l_s):
                for c in (s, w):
                    if not c == "":
                        ret.append(c)
            return ret, None
            #return [i[0] for i in ret if not i[0] == ""], None

    def pop_header(self, src_line):

        def pop_string(line):
            line = line.strip(" ")
            if " " in line:
                string, line = line.split(" ", 1)
                return string, line
            else:
                return line, ""

        def str2month(string):
            if not string in self.month_name:
                return None
            else:
                return self.month_name.index(string) + 1

        line = src_line[:]

        try:
            if self.re_datetime.match(line):
                # 2112-09-01 10:00:00 hostname mesasges...
                date_str, line = pop_string(line)
                time_str, line = pop_string(line)
                dt = datetime.datetime.strptime(" ".join((date_str, time_str)),
                        "%Y-%m-%d %H:%M:%S")
                host, line = pop_string(line)
                message = line
            else:
                # (2112) Sep 01 10:00:00 hostname messages...
                string, line = pop_string(line)
                if self.re_year.match(string):
                    year = int(string)
                    month_str, line = pop_string(line)
                    month = str2month(month_str)
                else:
                    year = self._set_year()
                    month = str2month(string)
                if month is None:
                    return None, None, None
                day_str, line = pop_string(line)
                day = int(day_str)
                time_str, line = pop_string(line)
                hour, minute, second = tuple(int(e) for e
                                             in time_str.split(":"))
                host, line = pop_string(line)
                message = line
                dt = datetime.datetime(year = year, month = month, day = day, 
                        hour = hour, minute = minute,
                        second = second, microsecond = 0)
        except:
            return None, None, None

        if self.header_size is not None:
            message = " ".join(src_line.split()[self.header_size:])
        return dt, host, message

    def process_line(self, line):
        line = line.rstrip("\n")
        if line == "":
            return None, None, None, None
        dt, host, message = self.pop_header(line)
        if message is None or message == "":
            return None, None, None, None
        elif self._is_removed(message):
            return None, None, None, None
        l_word, l_symbol = self.split_message(message)
        return dt, host, l_word, l_symbol


def iter_lines_log(conf, fn):
    lp = LogParser(conf)
    with open(fn, 'r') as f:
        for line in f:
            line = line.rstrip()
            dt, org_host, l_w, l_s = lp.process_line(line)
            if l_w is None: continue
            l_w = [strutil.add_esc(w) for w in l_w]
            yield line, l_w, l_s


def iter_lines_message(conf, fn):
    lp = LogParser(conf)
    with open(fn, 'r') as f:
        for line in f:
            line = line.rstrip()
            l_w, l_s = lp.split_message(line)
            if l_w is None: continue
            l_w = [strutil.add_esc(w) for w in l_w]
            yield line, l_w, l_s


def iter_lines(conf, fn, form):
    if form == "log":
        return iter_lines_log(conf, fn)
    elif form == "message":
        return iter_lines_message(conf, fn)
    else:
        raise NotImplementedError


def test_parse(conf):
    from . import common
    LP = LogParser(conf)
    ret = []
    if conf.getboolean("general", "src_recur"):
        l_fp = common.recur_dir(config.getlist(conf, "general", "src_path"))
    else:
        l_fp = common.rep_dir(config.getlist(conf, "general", "src_path"))
    for fp in l_fp:
        with open(fp, 'r') as f:
            for line in f:
                ret.append(LP.process_line(line.rstrip("\n")))
    return ret


if __name__ == "__main__":
    if len(sys.argv) < 3:
        sys.exit("usage: {0} config targets".format(sys.argv[0]))
    conf = config.open_config(sys.argv[1])
    
    from . import common
    LP = LogParser(conf)
    for fp in common.rep_dir(sys.argv[2:]):
        with open(fp) as f:
            for line in f:
                #print LP.process_line(line.rstrip("\n"))
                print(" ".join(LP.process_line(line.rstrip("\n"))[2]))





#!/usr/bin/env python
# coding: utf-8

"""
LenMa: Length Matters Clustering, proposed in [1].
K. Shima. Length Matters: Clustering System Log Messages using Length of Words. arXiv, 2016.

This code is based on https://github.com/keiichishima/templateminer .
"""

import numpy as np
from collections import defaultdict

from amulog import lt_common


class Cluster:

    def __init__(self, words):
        self._words = words
        self._nwords = words
        self._wordlens = np.array([len(w) for w in words])

    def _get_similarity_score_cosine(self, new_words):
        # cosine similarity
        from sklearn.metrics.pairwise import cosine_similarity
        wordlens = self._wordlens.reshape(1, -1)
        new_wordlens = np.asarray([len(w) for w in new_words]).reshape(1, -1)
        cos_score = cosine_similarity(wordlens, new_wordlens)
        return cos_score

    def get_similarity_score(self, new_words, n_same_count=3, use_head_rule=True):
        # heuristic judge: the first word (process name) must be equal
        if use_head_rule and not self._words[0] == new_words[0]:
            return 0.0
        cnt_wildcard = sum(1 for w in self._words if w == lt_common.REPLACER)
        cnt_same = sum(1 for w1, w2 in zip(self._nwords, new_words)
                       if w1 == w2)
        # check exact match
        if cnt_same + cnt_wildcard == self._nwords:
            return 1.0
        # heuristics?
        if cnt_same < n_same_count:
            return 0.0

        return self._get_similarity_score_cosine(new_words)

    def update(self, new_words):
        self._wordlens = np.array([len(w) for w in new_words])
        self._words = [w1 if w1 == w2 else lt_common.REPLACER
                       for w1, w2 in zip(self._words, new_words)]


class LTGenLenMa(lt_common.LTGen):

    def __init__(self, table, threshold=0.9, n_same_count=3, use_head_rule=True):
        super().__init__(table)
        self._threshold = threshold
        self._n_same_count = n_same_count
        self._use_head_rule = use_head_rule
        self._clusters = {}  # key: tid, val: Cluster
        self._d_candidates = defaultdict(list)  # key: length, val: list of tid

    def load(self, loadobj):
        self._clusters, self._d_candidates = loadobj

    def dumpobj(self):
        return self._clusters, self._d_candidates

    def process_line(self, pline):
        words = pline["words"]
        nwords = len(words)

        candidates = []
        for tid in self._d_candidates[nwords]:
            cluster = self._clusters[tid]
            score = cluster.get_similarity_score(words, self._n_same_count,
                                                 self._use_head_rule)
            if score >= self._threshold:
                candidates.append((tid, score))

        if len(candidates) > 0:
            tid = max(candidates, key=lambda x: x[1])[0]
            state = self.merge_tpl(words, tid)
            return tid, state
        else:
            tid = self.add_tpl(words)
            self._clusters[tid] = Cluster(words)
            self._d_candidates[nwords].append(tid)
            return tid, self.state_added


def init_ltgen(conf, table, **_):
    threshold = conf.getfloat("log_template_lenma", "threshold")
    n_same_count = conf.getint("log_template_lenma", "n_same_count")
    use_head_rule = conf.getboolean("log_template_lenma", "use_head_rule")

    return LTGenLenMa(table, threshold=threshold, n_same_count=n_same_count,
                      use_head_rule=use_head_rule)


def get_param_candidates():
    from itertools import product
    params = []
    for th, n_cnt in product((0., 0.5, 0.7, 0.8, 0.9, 0.95, 0.99), (2, 3, 4)):
        params.append({"threshold": th,
                       "n_same_count": n_cnt})
    return params


def init_ltgen_with_params(conf, table, params, **_):
    threshold = params["threshold"]
    n_same_count = params["n_same_count"]
    use_head_rule = conf.getboolean("log_template_lenma", "use_head_rule")

    return LTGenLenMa(table, threshold=threshold, n_same_count=n_same_count,
                      use_head_rule=use_head_rule)




"""
A system log management tool.
The main function is to classify log messages based
on their log templates (formats and variable locations),
and to store the data in a database.
"""

__version__ = '0.2.2'




#!/usr/bin/env python
# coding: utf-8

"""
Interface to use some functions about Log DB from CLI.
"""

import datetime
import logging

from . import cli
from . import common
from . import config

_logger = logging.getLogger(__package__)
SUBLIB = ["edit", "eval"]


def get_targets_arg(ns):
    if ns.recur:
        targets = common.recur_dir(ns.files)
    else:
        targets = common.rep_dir(ns.files)
    return targets


def get_targets_conf(conf):
    l_path = config.getlist(conf, "general", "src_path")
    if conf.getboolean("general", "src_recur"):
        return common.recur_dir(l_path)
    else:
        return common.rep_dir(l_path)


def get_targets(ns, conf):
    if ns is None or len(ns.files) == 0:
        return get_targets_conf(conf)
    else:
        return get_targets_arg(ns)


def is_online(conf, parallel):
    from .alg import is_online as alg_is_online
    mode = conf["log_template"]["processing_mode"]
    lt_methods = config.getlist(conf, "log_template", "lt_methods")
    return alg_is_online(mode, lt_methods, parallel)


def data_from_db(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)
    dirname = ns.dirname
    if ns.incr:
        method = "incremental"
    else:
        method = "commit"
    reset = ns.reset

    from . import log_db
    log_db.data_from_db(conf, dirname, method, reset)


def data_from_data(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)
    dirname = ns.dirname
    targets = get_targets(ns, conf)
    if ns.incr:
        method = "incremental"
    else:
        method = "commit"
    reset = ns.reset

    from . import manager
    manager.data_from_data(conf, targets, dirname, method, reset)


def data_parse(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)
    targets = get_targets(ns, conf)

    from . import strutil
    from . import manager
    lp = manager.load_log2seq(conf)
    for line in manager.iter_lines(targets):
        pline = manager.parse_line(strutil.add_esc(line), lp)
        print(pline)


def db_make(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)
    targets = get_targets(ns, conf)

    from . import manager
    timer = common.Timer("db-make", output=_logger)
    timer.start()
    if is_online(conf, ns.parallel):
        manager.process_files_online(conf, targets, True)
    else:
        manager.process_files_offline(conf, targets, True, ns.parallel)
    timer.stop()


def db_add(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)
    targets = get_targets_arg(ns)

    from . import manager

    timer = common.Timer("db-add", output=_logger)
    timer.start()
    if is_online(conf, ns.parallel):
        manager.process_files_online(conf, targets, False)
    else:
        manager.process_files_offline(conf, targets, False, ns.parallel)
    timer.stop()


def db_remake_group(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)

    from . import manager
    timer = common.Timer("db-remake-group", output=_logger)
    timer.start()
    manager.remake_ltgroup(conf)
    timer.stop()


def db_tag(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)

    from . import lt_label
    timer = common.Timer("db-tag", output=_logger)
    timer.start()
    lt_label.generate_all_tags(conf)
    timer.stop()


def db_anonymize(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)

    timer = common.Timer("db-anonymize", output=_logger)
    timer.start()
    if ns.conf_export:
        conf2 = config.open_config(ns.conf_export)
    else:
        conf2 = None
    from . import anonymize
    am = anonymize.AnonymizeMapper(conf)
    am.anonymize(conf2)
    fp = am.dump()
    timer.stop()

    print("> " + fp)


def db_anonymize_mapping(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)

    from . import anonymize
    am = anonymize.AnonymizeMapper(conf)
    am.mapping()
    fp = am.dump()

    print("> " + fp)


def show_db_info(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)
    from . import log_db

    log_db.info(conf)


def show_lt(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)
    from . import log_db

    simple = ns.simple
    print(log_db.show_all_lt(conf, simple=simple))


def show_ltg(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)
    from . import log_db

    print(log_db.show_all_ltg(conf))


def show_tag(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)
    from . import log_db

    print(log_db.show_tag(conf, tag=ns.tag))


def show_tag_stats(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)
    from . import log_db

    print(log_db.show_tag_stats(conf))


def show_lt_import(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)
    from . import log_db

    external = ns.external
    ld = log_db.LogData(conf)
    for ltobj in ld.iter_lt():
        if external:
            # unsegmented, with escape
            print(ltobj.restore_message(None, esc=True))
        else:
            # segmented, with escape
            print(" ".join(ltobj.ltw))


def show_host(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)
    from . import log_db

    log_db.show_all_host(conf)


def show_log(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)
    from . import log_db
    lidflag = ns.lid

    d = parse_condition(ns.conditions)
    ld = log_db.LogData(conf)
    for e in ld.iter_lines(**d):
        if lidflag:
            print("{0} {1}".format(e.lid, e.restore_line()))
        else:
            print(e.restore_line())


def parse_condition(conditions):
    """
    Args:
        conditions (list)
    """
    d = {}
    for arg in conditions:
        if "=" not in arg:
            raise SyntaxError
        key = arg.partition("=")[0]
        if key == "ltid":
            d["ltid"] = int(arg.partition("=")[-1])
        elif key in ("gid", "ltgid"):
            d["ltgid"] = int(arg.partition("=")[-1])
        elif key in ("date_from", "top_date"):
            date_string = arg.partition("=")[-1]
            d["dts"] = datetime.datetime.strptime(date_string, "%Y-%m-%d")
        elif key in ("date_to", "end_date"):
            date_string = arg.partition("=")[-1]
            d["dte"] = datetime.datetime.strptime(date_string, "%Y-%m-%d")
        elif key == "date":
            date_string = arg.partition("=")[-1]
            d["dts"] = datetime.datetime.strptime(date_string, "%Y-%m-%d")
            d["dte"] = d["dts"] + datetime.timedelta(days=1)
        elif key == "host":
            d["host"] = arg.partition("=")[-1]
    return d


def conf_defaults(_):
    config.show_default_config()


def conf_diff(ns):
    files = ns.files[:]
    if ns.configset:
        files += config.read_config_group(ns.configset)
    config.show_config_diff(files)


def conf_minimum(ns):
    conf = config.open_config(ns.conf_path, base_default=False, ignore_import=True)
    conf = config.minimize(conf)
    config.write(ns.conf_path, conf)
    print("rewrite {0}".format(ns.conf_path))


def conf_set_edit(ns):
    l_conf_name = config.read_config_group(ns.configset)
    key = ns.key
    rulestr = ns.rule

    if "(" in rulestr:
        temp = rulestr.rstrip(")").split("(")
        if not len(temp) == 2:
            raise ValueError("bad format for value specification")
        rulename, argstr = temp
        args = argstr.split(",")

        if rulename == "list":
            assert len(args) == len(l_conf_name)
            d = {key: args}
        elif rulename == "range":
            assert len(args) == 2
            start, step = [int(v) for v in args]
            l_val = [start + i * step for i in range(len(l_conf_name))]
            d = {key: l_val}
        elif rulename == "power":
            assert len(args) == 2
            start, step = [int(v) for v in args]
            l_val = [start * (i ** step) for i in range(len(l_conf_name))]
            d = {key: l_val}
        elif rulename == "withconf":
            assert len(args) == 1
            l_val = [args[0] + name for i, name in enumerate(l_conf_name)]
            d = {key: l_val}
        elif rulename == "namerange":
            assert len(args) == 1
            l_val = [args[0] + str(i) for i in range(len(l_conf_name))]
            d = {key: l_val}
        else:
            raise NotImplementedError("invalid rule name")
    else:
        l_val = [rulestr] * len(l_conf_name)
        d = {key: l_val}

    config.config_group_edit(l_conf_name, d)


def conf_shadow(ns):
    cond = {}
    incr = []
    for rule in ns.rules:
        if "=" in rule:
            key, val = rule.split("=")
            cond[key] = val
        else:
            incr.append(rule)
    l_conf_name = config.config_shadow(n=ns.number, cond=cond, incr=incr,
                                       fn=ns.conf_path, output=ns.output,
                                       ignore_overwrite=ns.force)

    if ns.configset is not None:
        config.dump_config_group(ns.configset, l_conf_name)
        print(ns.configset)


# common argument settings
OPT_DEBUG = [["--debug"],
             {"dest": "debug", "action": "store_true",
              "help": "set logging level to debug (default: info)"}]
OPT_CONFIG = [["-c", "--config"],
              {"dest": "conf_path", "metavar": "CONFIG", "action": "store",
               # "default": config.DEFAULT_CONFIG,
               "default": None,
               "help": "configuration file path for amulog"}]
OPT_CONFIG_SET = [["-s", "--configset"],
                  {"dest": "configset", "metavar": "CONFIG_SET",
                   "default": None,
                   "help": "use config group definition file"}]
OPT_PARALLEL = [["-p", "--parallel"],
                {"dest": "parallel", "action": "store_true",
                 "help": "parallel processing in offline mode"}]
OPT_RECUR = [["-r", "--recur"],
             {"dest": "recur", "action": "store_true",
              "help": "recursively search files to process"}]
OPT_DRY = [["-d", "--dry"],
           {"dest": "dry", "action": "store_true",
            "help": "do not store data into db"}]
# OPT_TERM = [["-t", "--term"],
#             {"dest": "dt_range",
#              "metavar": "DATE1 DATE2", "nargs": 2,
#              "help": ("datetime range, start and end in YY-MM-dd style. "
#                       "(optional; defaultly use all data)")}]
ARG_FILE = [["file"],
            {"metavar": "PATH", "action": "store",
             "help": "filepath to output"}]
ARG_FILES = [["files"],
             {"metavar": "PATH", "nargs": "+",
              "help": "files or directories as input"}]
ARG_FILES_OPT = [["files"],
                 {"metavar": "PATH", "nargs": "*",
                  "help": ("files or directories as input "
                           "(optional; defaultly read from config")}]
ARG_DBSEARCH = [["conditions"],
                {"metavar": "CONDITION", "nargs": "+",
                 "help": ("Conditions to search log messages. "
                          "Example: MODE gid=24 date=2012-10-10 ..., "
                          "Keys: ltid, gid, date, date_from, date_to, host")}]

# argument settings for each modes
# description, List[args, kwargs], func
# defined after functions because these settings use functions
DICT_ARGSET = {
    "data-from-db": ["Generate log data from DB.",
                     [OPT_CONFIG, OPT_DEBUG,
                      [["-d", "--dirname"],
                       {"dest": "dirname", "metavar": "DIRNAME",
                        "action": "store",
                        "help": "directory name to output"}],
                      [["-i", "--incr"],
                       {"dest": "incr", "action": "store_true",
                        "help": "output incrementally, use with small memory"}],
                      [["--reset"],
                       {"dest": "reset", "action": "store_true",
                        "help": "reset log file directory before processing"}],
                      ],
                     data_from_db],
    "data-from-data": ["Re-arrange log file, splitting messages by date.",
                       [OPT_CONFIG, OPT_DEBUG, OPT_RECUR, ARG_FILES_OPT,
                        [["-d", "--dirname"],
                         {"dest": "dirname", "metavar": "DIRNAME",
                          "action": "store",
                          "help": "directory name to output"}],
                        [["-i", "--incr"],
                         {"dest": "incr", "action": "store_true",
                          "help": "output incrementally, use with small memory"}],
                        [["--reset"],
                         {"dest": "reset", "action": "store_true",
                          "help": "reset log file directory before processing"}],
                        ],
                       data_from_data],
    "data-parse": ["Check log data parsing with log2seq.",
                   [OPT_CONFIG, OPT_DEBUG, OPT_RECUR, ARG_FILES_OPT],
                   data_parse],
    "db-make": ["Initialize database and add log data. ",
                [OPT_CONFIG, OPT_DEBUG, OPT_RECUR, OPT_PARALLEL, ARG_FILES_OPT],
                db_make],
    "db-add": ["Add log data to existing database.",
               [OPT_CONFIG, OPT_DEBUG, OPT_RECUR, OPT_PARALLEL, ARG_FILES],
               db_add],
    "db-remake-group": ["Remake log template groups",
                        [OPT_CONFIG, OPT_DEBUG],
                        db_remake_group],
    "db-tag": ["Make log template tags.",
               [OPT_CONFIG, OPT_DEBUG],
               db_tag],
    "db-anonymize": ["Anonymize templates and hostnames.",
                     [OPT_CONFIG, OPT_DEBUG,
                      [["--config-export"],
                       {"dest": "conf_export", "metavar": "CONFIG_EXPORT",
                        "action": "store", "default": None,
                        "help": "generate another DB with given config"}],
                      ],
                     db_anonymize],
    "db-anonymize-mapping": ["Generate anonymization mapping json.",
                             [OPT_CONFIG, OPT_DEBUG],
                             db_anonymize_mapping],
    "show-db-info": ["Show abstruction of database status.",
                     [OPT_CONFIG, OPT_DEBUG],
                     show_db_info],
    "show-lt": ["Show all log templates in database.",
                [OPT_CONFIG, OPT_DEBUG,
                 [["-s", "--simple"],
                  {"dest": "simple", "action": "store_true",
                   "help": "only show log templates"}]],
                show_lt],
    "show-ltg": ["Show all log template groups and their members.",
                 [OPT_CONFIG, OPT_DEBUG,
                  [["-g", "--group"],
                   {"dest": "group", "action": "store", "default": None,
                    "help": "show members of given labeling group"}]],
                 show_ltg],
    "show-tag": ["Show all tags and their templates.",
                 [OPT_CONFIG, OPT_DEBUG,
                  [["tag"],
                   {"metavar": "TAG", "action": "store",
                    "nargs": "?", "default": None,
                    "help": "tag to search (if not given, show all tags)"}]],
                 show_tag],
    "show-tag-stats": ["Show all tags and their templates.",
                       [OPT_CONFIG, OPT_DEBUG],
                       show_tag_stats],
    "show-lt-import": ["Output log template definitions in importable format.",
                       [OPT_CONFIG, OPT_DEBUG,
                        [["-e", "--external"],
                         {"dest": "external", "action": "store_true",
                          "help": "output in non-amulog RE parser format"}]],
                       show_lt_import],
    "show-host": ["Show all hostnames in database.",
                  [OPT_CONFIG, OPT_DEBUG],
                  show_host],
    "show-log": ["Search and show log messages.",
                 [OPT_CONFIG, OPT_DEBUG,
                  [["--lid"],
                   {"dest": "lid", "action": "store_true",
                    "help": "show lid"}],
                  ARG_DBSEARCH],
                 show_log],
    "conf-defaults": ["Show default configurations.",
                      [],
                      conf_defaults],
    "conf-diff": ["Show differences of 2 configuration files.",
                  [OPT_CONFIG_SET,
                   [["files"],
                    {"metavar": "FILENAME", "nargs": "*",
                     "help": "configuration file"}]],
                  conf_diff],
    "conf-minimum": ["Remove default options and comments.",
                     [[["-o", "--overwrite"],
                       {"dest": "overwrite", "action": "store_true",
                        "help": "overwrite file instead of stdout dumping"}],
                      [["conf_path"],
                       {"metavar": "PATH",
                        "help": "config filepath to load"}]],
                     conf_minimum],
    "conf-group-edit": ["Edit configuration files in a config group.",
                        [[["configset"],
                          {"metavar": "CONFIG_SET",
                           "help": "config group definition file to use"}],
                         [["key"],
                          {"metavar": "KEY",
                           "help": "\"SECTION.OPTION\" to edit"}],
                         [["rule"],
                          {"metavar": "RULE",
                           "help": ("Value specification rule "
                                    "defined in function-like format. "
                                    "Example: \"List(1,10,100,1000)\" "
                                    "Available format: "
                                    "list(values for each config), "
                                    "range(START,STEP), "
                                    "power(START,STEP), "
                                    "withconf(NAME),"
                                    "namerange(NAME)."
                                    "Note that 1 rule for 1 execution.")}]],
                        conf_set_edit],
    "conf-shadow": ["Copy configuration files.",
                    [OPT_CONFIG,
                     [["-f", "--force"],
                      {"dest": "force", "action": "store_true",
                       "help": "Ignore overwrite of output file"}],
                     [["-n", "--number"],
                      {"dest": "number", "metavar": "INT",
                       "action": "store", "type": int, "default": 1,
                       "help": "number of files to generate"}],
                     [["-o", "--output"],
                      {"dest": "output", "metavar": "FILENAME",
                       "action": "store", "type": str, "default": None,
                       "help": "basic output filename"}],
                     [["-s", "--configset"],
                      {"dest": "configset", "metavar": "CONFIG_SET",
                       "default": None,
                       "help": ("define config group "
                                "and dump it in given filename")}],
                     [["rules"],
                      {"metavar": "RULES", "nargs": "*",
                       "help": ("Rules to replace options. You can indicate "
                                "option, or option and its value with =. "
                                "You can use both of them together. "
                                "For example: \"general.import=hoge.conf "
                                "general.logging\"")}]],
                    conf_shadow],
}

ALIASES = {
    "dump-lt": "show-lt-import",
}


def main():
    cli.main(DICT_ARGSET, ALIASES, SUBLIB)


if __name__ == "__main__":
    main()

    # import cProfile
    # cProfile.run('_main()', filename='main.prof')



#!/usr/bin/env python
# coding: utf-8

import ipaddress
from collections import defaultdict

from . import config


class HostAlias(object):
    """
    Note:
        1 host can not belong to multiple host groups, because
        group definition is used to label and classify variables
        in log templates.
    """

    def __init__(self, fn):
        self._fn = fn
        self._d_alias = defaultdict(list)  # key = alias, val = List[host]
        self._d_ralias = {}  # key = host, val = alias
        self._d_group = defaultdict(list)  # key = group, val = List[host]
        self._d_rgroup = {}  # key = host, val = group
        self._l_net = []
        self._open(self._fn)

    def _open(self, fn):
        group = "default"
        if fn is None or fn == "":
            return
        with open(fn, "r") as f:
            for line in f:
                line = line.rstrip("\n")
                line = line.partition("#")[0]
                if line == "" or line[0] == "#":
                    continue
                elif line[0] == "[" and "]" in line:
                    group = line.strip("[").partition("]")[0]
                elif line[0] == "<" and ">" in line:
                    l_temp = line.strip("<").partition(">")
                    alias = l_temp[0]
                    names = [alias] + l_temp[2].strip().rstrip("\n").split()
                    self._add_def(names, alias=alias, group=group)
                else:
                    names = line.rstrip("\n").split()
                    if len(names) == 0:
                        continue
                    self._add_def(names, group=group)

    def _add_def(self, l_name, alias=None, group=None):

        def add_alias(name, alias):
            if alias is not None:
                self._d_alias[alias].append(name)
                self._d_ralias[name] = alias
            else:
                self._d_alias[name].append(name)
                self._d_ralias[name] = name

        def add_groupdef(key, group):
            if group is not None:
                self._d_group[group].append(key)
                self._d_rgroup[key] = group

        for name in l_name:
            if "/" in name:
                try:
                    net = str(ipaddress.ip_network(name))
                    add_alias(net, alias)
                    add_groupdef(net, group)
                    self._l_net.append(net)
                except ValueError:
                    add_alias(name, alias)
                    add_groupdef(name, group)
            else:
                try:
                    addr = str(ipaddress.ip_address(name))
                    add_alias(addr, alias)
                    add_groupdef(addr, group)
                except ValueError:
                    add_alias(name, alias)
                    add_groupdef(name, group)

    def keys(self):
        return self._d_alias.keys()

    def print_definitions(self):
        print("[aliases]")
        for k, val in self._d_alias.items():
            print("<{0}> ".format(k) + " ".join([str(v) for v in val]))
        print()
        print("[groups]")
        for k, val in self._d_group.items():
            print(k)
            print(" ".join([str(v) for v in val]))
            print()

    def isknown(self, string):
        try:
            addr = str(ipaddress.ip_address(string))
            for net in self._l_net:
                if addr in net:
                    return True
            else:
                if addr in self._d_ralias:
                    return True
                else:
                    return False
        except ValueError:
            name = string.lower()
            return name in self._d_ralias

    def resolve_host(self, string):
        try:
            addr = str(ipaddress.ip_address(string))
            for net in self._l_net:
                if addr in net:
                    return self._d_ralias[net]
            else:
                if addr in self._d_ralias:
                    return self._d_ralias[addr]
                else:
                    return None
        except ValueError:
            name = string.lower()
            if name in self._d_ralias:
                return self._d_ralias[name]
            else:
                return None

    def group(self, group):
        return self._d_group[group]

    def get_group(self, string):
        try:
            addr = str(ipaddress.ip_address(string))
            for net in self._l_net:
                if addr in net:
                    return self._d_rgroup[net]
            else:
                if addr in self._d_rgroup:
                    return self._d_rgroup[addr]
                else:
                    return None
        except ValueError:
            name = string.lower()
            if name in self._d_rgroup:
                return self._d_rgroup[name]
            else:
                return None


def init_hostalias(conf):
    ha_fn = conf["manager"]["host_alias_filename"]
    ha = HostAlias(ha_fn)
    return ha


def test_hostalias(conf):
    names = ["192.168.0.1",
             "www.TEST.localdomain",
             "localhost",
             "www",
             "www3",
             "hoge",
             "10.100.1.254",
             "8.8.6.0"]
    # conf.set("database", "host_alias_filename", "host_alias_test.txt")
    ha = init_hostalias(conf)
    # ha = HostAlias(conf)
    ha.print_definitions()
    print()
    print("[test aliasing]")
    for name in names:
        print(name)
        print(ha.isknown(name))
        print(ha.resolve_host(name))
        print(ha.get_group(name))
        print()


#if __name__ == "__main__":
#    usage = ""
#    import optparse
#
#    op = optparse.OptionParser(usage)
#    op.add_option("-c", "--config", action="store",
#                  dest="conf", type="string", default=config.DEFAULT_CONFIG_NAME,
#                  help="configuration file path")
#    options, args = op.parse_args()
#    conf = config.open_config(options._conf)
#    test_hostalias(conf)



#!/usr/bin/env python
# coding: utf-8

"""
Construct and manage database of log messages, templates,
and grouping definitions.
"""

import datetime
import logging
from collections import defaultdict

from . import common
from . import strutil
from . import db_common
from . import lt_common

_logger = logging.getLogger(__package__)


class LogMessage:
    """An annotated log message.

    An instance have a set of information about 1 log message line,
    including timestamp, hostname, log template, and message contents.

    Attributes:
        lid (int): A message identifier in DB.
        lt (lt_common.LogTemplate): Log template of this message.
        dt (datetime.datetime): A timestamp for this message.
        host (str): A hostname that output this message.
        l_w (List(str)): A sequence of words in this message.

    """

    def __init__(self, lid, lt, dt, host, l_w):
        """
        Args:
            lid (int): A message identifier in DB.
            lt (lt_common.LogTemplate): Log template of this message.
            dt (datetime.datetime): A timestamp for this message.
            host (str): A hostname that output this message.
            l_w (List(str)): A sequence of words in this message.

        """
        self.lid = lid
        self.lt = lt
        self.dt = dt
        self.host = host
        self.l_w = l_w

    def __str__(self):
        """str: Show attributes in 1 string."""
        return " ".join((str(self.dt), self.host, str(self.lt.ltid),
                         str(self.l_w)))

    def get(self, name):
        """Return value for given attribute."""
        if name == "lid":
            return self.lid
        elif name == "ltid":
            return self.lt.ltid
        elif name == "ltgid":
            return self.lt.ltgid
        elif name == "host":
            return self.host
        else:
            raise NotImplementedError

    def var(self):
        """str: Get sequence of all variable words in this message.
        Variable words mean what is presented with mask
        (defaults **) in log template.
        """
        return self.lt.var(self.l_w)

    def restore_message(self):
        """str: Get pseudo original log message
        except headers (timestamp and hostname).
        """
        return self.lt.restore_message(self.l_w)

    def restore_line(self):
        """str: Get original log message contents
        including headers (timestamp and hostname).
        """
        return " ".join((str(self.dt), str(self.host),
                         self.restore_message()))


class LogData:
    """Interface to get, add or edit log messages in DB.

    Attributes:
        conf (config.ExtendedConfigParser): A common configuration object.
        lttable (lt_common.LTTable): Log template table object.
        db (LogDB): Log database instance.
    """

    def __init__(self, conf, edit=False, reset_db=False):
        """
        Args:
            conf (config.ExtendedConfigParser): A common configuration object.
            edit (Optional[bool]): Defaults to False.
                True if database will be added or edited.
                False if database is used in readonly mode.
            reset_db (Optional[bool]): Defaults to False.
                If True, database will be reset before following process.

        """
        self.conf = conf
        self._reset_db = reset_db

        self.db = LogDB(conf, edit, reset_db)
        self.lttable = None
        if edit:
            if reset_db:
                # use empty lttable
                self.lttable = lt_common.LTTable()
            else:
                # load existing lttable to edit
                self.lttable = self.db.restore_lttable()
        else:
            # load existing lttable to read
            self.lttable = self.db.restore_lttable()

    def _row_to_lm(self, d_line):
        ltid = d_line.pop("ltid")
        d_line["lt"] = self.lttable[ltid]
        return LogMessage(**d_line)

    def add_line(self, lid, ltid, dt, host, l_w):
        """Directly add LogMessage to DB.
        Used in Anonymize functions."""
        self.db.add_line(lid=lid, ltid=ltid, dt=dt, host=host, l_w=l_w)

    def add_lt(self, ltobj):
        """Directly add LogTemplate to DB.
        Used in Anonymize functions."""
        self.db.add_lt(ltobj)
        self.db.add_ltg(ltobj.ltid, ltobj.ltgid)

    def get_line(self, lid):
        return self._row_to_lm(self.db.get_line(lid))

    def iter_lines(self, **kwargs):
        """Generate log messages in DB that satisfy conditions
        given in arguments. All arguments are defaults to None, and ignored.

        Keyword Args:
            lid (int): A message identifier in DB.
            ltid (int): A log template identifier.
            ltgid (int): A log template grouping identifier.
            dts (datetime.datetime): Condition for timestamp.
                Messages after 'dts' will be yield.
            dte (datetime.datetime): Condition for timestamp.
                Messages before 'dte' will be yield.
            host (str): A source hostname of the message.

        Yields:
            LogMessage: An annotated log message instance
                which satisfies all given conditions.
        """
        _logger.debug("iter_lines called ({0})".format(" ".join(
            ["{0}:{1}".format(k, v) for k, v in kwargs.items()
             if v is not None])
        ))
        assert len(kwargs) >= 1, "empty arguments"
        for d_line in self.db.iter_lines(kwargs):
            yield self._row_to_lm(d_line)

    def iter_all(self):
        for d_line in self.db.iter_all():
            yield self._row_to_lm(d_line)

    def get_tags(self, **kwargs):
        """Search tags for given template identifiers.
        One of the args should be given.

        Keyword Args:
            ltid (Optional[int]): Search tags for given log template identifier.
            ltgid (Optional[int]): Search tags for given group identifier.

        Returns:
            List[str]
        """
        assert len(kwargs) >= 1, "empty arguments"
        return self.db.get_tags(**kwargs)

    def count_lines(self):
        """int: Number of all messages in DB."""
        return self.db.count_lines()

    def dt_term(self):
        """datetime.datetime, datetime.datetime:
            Period of all registered log messages in DB."""
        return self.db.dt_term()

    def whole_term(self):
        """datetime.datetime, datetime.datetime:
            Period of dates that include all log messages in DB.
        """
        # return date term that cover all log data
        top_dt, end_dt = self.db.dt_term()
        top_dt = datetime.datetime.combine(
            top_dt.date(), datetime.time()
        )
        end_dt = datetime.datetime.combine(
            end_dt.date(), datetime.time()
        ) + datetime.timedelta(days=1)
        return top_dt, end_dt

    def whole_host_lt(self, dts=None, dte=None):
        """List[str, str]: Sequence of all combinations of
        hostname and ltids in DB."""
        return self.db.whole_host_lt(dts=dts, dte=dte)

    def whole_host_ltg(self, dts=None, dte=None):
        """List[str, str]: Sequence of all combinations of
        hostname and ltgids in DB."""
        ret = set()
        for host, ltid in self.whole_host_lt(dts=dts, dte=dte):
            ret.add((host, self.ltgid_from_ltid(ltid)))
        return list(set(ret))

    def whole_host(self, dts=None, dte=None):
        """List[str]: Sequence of all source hostname in DB."""
        return self.db.whole_host(dts=dts, dte=dte)

    def count_lt(self):
        """int: Number of all log templates."""
        return self.db.count_lt()

    def count_ltg(self):
        """int: Number of all log template groups."""
        return self.db.count_ltg()

    def iter_lt(self):
        """Yields lt_common.LogTemplate: All log template instance in DB."""
        for ltline in self.lttable:
            yield ltline

    def lt(self, ltid):
        """Get log template instance of given log template identifier.

        Args:
            ltid (int): A log template identifier.

        Returns:
            lt_common.LogTemplate: A log template instance.

        """
        return self.lttable[ltid]

    def iter_gid(self, gid_name):
        if gid_name == "ltid":
            return [lt.ltid for lt in self.iter_lt()]
        elif gid_name == "ltgid":
            return self.iter_ltgid()

    def iter_ltgid(self):
        """Yields int: Get all identifiers of log template groups."""
        return self.db.iter_ltgid()

    def ltgid_from_ltid(self, ltid):
        lt = self.lttable[ltid]
        return lt.ltgid

    def ltg_members(self, ltgid):
        """Get all log templates in given log template group.

        Args:
            ltgid (int): A log template group identifier.

        Returns:
            List[lt_common.LogTemplates]: Sequence of all log template
                instances that belongs to given log template group.

        """
        return [self.lttable[ltid] for ltid in self.db.get_ltg_members(ltgid)]

    def show_lt_info(self, ltid):
        ltobj = self.lt(ltid)
        tags = self.get_tags(ltid=ltobj.ltid)
        if tags:
            tag_str = "|".join(tags)
            return " ".join((str(ltobj.ltid), "({0})".format(tag_str),
                             str(ltobj), "({0})".format(ltobj.count)))
        else:
            return " ".join((str(ltobj.ltid),
                             str(ltobj), "({0})".format(ltobj.count)))

    def show_ltg_info(self, ltgid):
        tags = self.get_tags(ltgid=ltgid)
        l_ltobj = self.ltg_members(ltgid)
        length = len(l_ltobj)
        count = sum(ltobj.count for ltobj in l_ltobj)

        header = "[ltgroup {0} ({1} tpls, {2} lines)]".format(
            ltgid, length, count, ", ".join(tags)
        )
        if tags:
            header += " # tags: {0}".format(",".join(tags))
        buf = [header]
        for ltobj in l_ltobj:
            buf.append(self.show_lt_info(ltobj.ltid))
        return "\n".join(buf)

    def commit_db(self):
        """Commit requested changes in LogDB.
        """
        self.db.commit()

    def drop_all(self):
        self.db.drop_all()


class LogDB:
    """Interface of DB transaction for log data.

    Note:
        It is not recommended to use this class directly.
        Instead, use LogData.
    """

    tablename_log = "log"
    tablename_lt = "lt"
    tablename_ltg = "ltg"
    tablename_tag = "tag"
    table_names = (tablename_log, tablename_lt, tablename_ltg, tablename_tag)
    indexname_log = "log_index"
    indexname_ltg = "ltg_index"
    indexname_tag = "tag_index"
    _tablename_tmp_footer = "_tmp"

    def __init__(self, conf, edit, reset_db):
        self._line_cnt = 0
        self._splitter = conf.get("database", "split_symbol")
        self._table_switch = {}

        db_type = conf.get("database", "database")
        if db_type == "sqlite3":
            from . import db_sqlite
            self._db = db_sqlite.init_db_conn(conf)
        elif db_type == "mysql":
            from . import db_mysql
            self._db = db_mysql.init_db_conn(conf)
        else:
            raise ValueError("invalid database type ({0})".format(
                db_type))

        if self._db.db_exists():
            if edit:
                if reset_db:
                    # create mode
                    _logger.info("DB reset")
                    self._db.reset()
                    self._init_tables()
                else:
                    # append mode
                    # load _line_cnt as next lid
                    self._line_cnt = self.count_lines()
            else:
                # read-only mode
                pass
        else:
            if edit:
                # create mode
                if reset_db:
                    msg = "Requested to reset DB, but database not found"
                    _logger.warning(msg)
                self._init_tables()
            else:
                raise IOError("database not found")

    def _init_tables(self):
        self._init_table_log()
        self._init_table_lt()
        self._init_table_ltg()
        self._init_table_tag()
        self._init_index()

    def _init_table_log(self, table_name=None):
        if table_name is None:
            table_name = self.tablename_log

        l_key = [db_common.TableKey("lid", "integer",
                                    # ("primary_key", "auto_increment", "not_null")),
                                    ("primary_key", "not_null")),
                 db_common.TableKey("ltid", "integer", tuple()),
                 db_common.TableKey("dt", "datetime", tuple()),
                 db_common.TableKey("host", "text", tuple()),
                 db_common.TableKey("words", "text", tuple())]
        sql = self._db.create_table_sql(table_name, l_key)
        self._db.execute(sql)

    def _init_table_lt(self, table_name=None):
        if table_name is None:
            table_name = self.tablename_lt
        l_key = [db_common.TableKey("ltid", "integer", ("primary_key",)),
                 db_common.TableKey("ltw", "text", tuple()),
                 db_common.TableKey("lts", "text", tuple()),
                 db_common.TableKey("count", "integer", tuple())]
        sql = self._db.create_table_sql(table_name, l_key)
        self._db.execute(sql)

    def _init_table_ltg(self, table_name=None):
        if table_name is None:
            table_name = self.tablename_ltg
        l_key = [db_common.TableKey("ltid", "integer", ("primary_key",)),
                 db_common.TableKey("ltgid", "integer", tuple())]
        sql = self._db.create_table_sql(table_name, l_key)
        self._db.execute(sql)

    def _init_table_tag(self, table_name=None):
        if table_name is None:
            table_name = self.tablename_tag
        l_key = [db_common.TableKey("ltid", "integer", tuple()),
                 db_common.TableKey("tag", "text", tuple())]
        sql = self._db.create_table_sql(table_name, l_key)
        self._db.execute(sql)

    def _init_index(self):
        self._init_index_log()
        self._init_index_ltg()
        self._init_index_tag()

    def _init_index_log(self):
        index_name = self.indexname_log
        table_name = self.tablename_log
        l_key = [db_common.TableKey("lid", "integer", tuple()),
                 db_common.TableKey("ltid", "integer", tuple()),
                 db_common.TableKey("dt", "datetime", tuple()),
                 db_common.TableKey("host", "text", (100,))]
        sql = self._db.create_index_sql(table_name, index_name, l_key)
        self._db.execute(sql)

    def _init_index_ltg(self):
        index_name = self.indexname_ltg
        table_name = self.tablename_ltg
        l_key = [db_common.TableKey("ltgid", "integer", tuple()), ]
        sql = self._db.create_index_sql(table_name, index_name, l_key)
        self._db.execute(sql)

    def _init_index_tag(self):
        index_name = self.indexname_tag
        table_name = self.tablename_tag
        l_key = [db_common.TableKey("tag", "text", (100,)), ]
        sql = self._db.create_index_sql(table_name, index_name, l_key)
        self._db.execute(sql)

    def switch_temporal_table(self, table_name):
        """Switch valid table to temporal one.
        Used for large data update."""
        tmp_table_name = table_name + self._tablename_tmp_footer
        assert tmp_table_name not in self._db.get_table_names()

        if table_name == self.tablename_log:
            self._init_table_log(tmp_table_name)
        elif table_name == self.tablename_lt:
            self._init_table_lt(tmp_table_name)
        elif table_name == self.tablename_ltg:
            self._init_table_ltg(tmp_table_name)
        elif table_name == self.tablename_tag:
            self._init_table_tag(tmp_table_name)
        else:
            raise ValueError("invalid table name {0}".format(table_name))

        self._table_switch[table_name] = tmp_table_name

    def apply_temporal_table(self, table_name):
        """Use temporal table as the main table.
        This function will remove the old main table,
        and rename temporal one as the main table.
        Used for large data update."""
        tmp_table_name = table_name + self._tablename_tmp_footer
        assert tmp_table_name in self._db.get_table_names()

        if table_name == self.tablename_log:
            index_name = self.indexname_log
            index_func = self._init_index_log
        elif table_name == self.tablename_ltg:
            index_name = self.indexname_ltg
            index_func = self._init_index_ltg
        elif table_name == self.tablename_tag:
            index_name = self.indexname_tag
            index_func = self._init_index_tag
        else:
            index_name = None
            index_func = None

        if index_name:
            sql = self._db.drop_index_sql(index_name)
            self._db.execute(sql)

        sql = self._db.drop_table_sql(table_name)
        self._db.execute(sql)

        sql = self._db.alter_table_sql(tmp_table_name, table_name)
        self._db.execute(sql)

        if index_name:
            index_func()

        self._table_switch.pop(table_name)

    def _valid_table_name(self, table_name):
        # for write functions only
        if table_name in self._table_switch:
            return self._table_switch[table_name]
        else:
            return table_name

    def commit(self):
        self._db.commit()

    def _parse_input(self, **kwargs):
        d = {}
        for k, v in kwargs.items():
            if k == "dt":
                d[k] = self._db.strftime(v)
            elif k == "l_w":
                d["words"] = self._splitter.join(v)
            elif k in ("lid", "ltid", "host"):
                d[k] = v
            else:
                raise KeyError
        return d

    def _parse_row(self, row):
        d_line = {"lid": int(row[0]),
                  "ltid": int(row[1]),
                  "dt": self._db.datetime(row[2]),
                  "host": row[3]}
        if row[4] == "":
            d_line["l_w"] = []
        else:
            d_line["l_w"] = strutil.split_igesc(row[4], self._splitter)
        return d_line

    # def add_line(self, ltid, dt, host, l_w, lid=None):
    def add_line(self, **kwargs):
        assert "ltid" in kwargs
        assert "dt" in kwargs
        assert "host" in kwargs
        assert "l_w" in kwargs
        d_val = self._parse_input(**kwargs)

        self._line_cnt += 1
        if "lid" not in kwargs:
            d_val["lid"] = self._line_cnt

        table_name = self._valid_table_name(self.tablename_log)
        l_ss = [db_common.StateSet(k, k) for k in d_val.keys()]
        sql = self._db.insert_sql(table_name, l_ss)
        self._db.execute(sql, d_val)

        return d_val["lid"]

    def iter_all(self):
        l_order = [("lid", "asc")]
        for row in self._select_log({}, l_order=l_order):
            d_line = {"lid": int(row[0]),
                      "ltid": int(row[1]),
                      "dt": self._db.datetime(row[2]),
                      "host": row[3]}
            if row[4] == "":
                d_line["l_w"] = []
            else:
                d_line["l_w"] = strutil.split_igesc(row[4], self._splitter)
            yield d_line

    def iter_lines(self, conditions):
        d_cond = {k: v for k, v in conditions.items()
                  if v is not None}
        if len(d_cond) == 0:
            raise ValueError("Use iter_all to get all lines")

        # for compatibility
        if "top_dt" in d_cond and "dts" not in d_cond:
            d_cond["dts"] = d_cond.pop("top_dt")
        if "end_dt" in d_cond and "dte" not in d_cond:
            d_cond["dte"] = d_cond.pop("end_dt")

        for row in self._select_log(d_cond):
            yield self._parse_row(row)

    def iter_words(self, conditions):
        d_cond = {k: v for k, v in conditions.items()
                  if v is not None}
        if len(d_cond) == 0:
            raise ValueError("Use iter_all to get all lines")

        # for compatibility
        if "top_dt" in d_cond and "dts" not in d_cond:
            d_cond["dts"] = d_cond.pop("top_dt")
        if "end_dt" in d_cond and "dte" not in d_cond:
            d_cond["dte"] = d_cond.pop("end_dt")

        for row in self._select_log(d_cond):
            if row[4] == "":
                yield []
            else:
                yield strutil.split_igesc(row[4], self._splitter)

    def _select_log(self, d_cond, l_order=None):
        # if len(d_cond) == 0:
        #     raise ValueError("called select with empty condition")
        args = d_cond.copy()

        table_name = self.tablename_log
        l_key = ["lid", "ltid", "dt", "host", "words"]
        l_cond = []
        for c in d_cond.keys():
            if c == "ltgid":
                sql = self._db.select_sql("ltg", ["ltid"],
                                          [db_common.Condition(c, "=", c, True)])
                l_cond.append(db_common.Condition("ltid", "in", sql, False))
            elif c == "dts":
                l_cond.append(db_common.Condition("dt", ">=", c, True))
                args[c] = self._db.strftime(d_cond[c])
            elif c == "dte":
                l_cond.append(db_common.Condition("dt", "<", c, True))
                args[c] = self._db.strftime(d_cond[c])
            else:
                l_cond.append(db_common.Condition(c, "=", c, True))
        sql = self._db.select_sql(table_name, l_key, l_cond, l_order)
        return self._db.execute(sql, args)

    def get_line(self, lid):
        table_name = self.tablename_log
        l_key = ["lid", "ltid", "dt", "host", "words"]
        l_cond = [db_common.Condition("lid", "=", "lid", True)]
        sql = self._db.select_sql(table_name, l_key, l_cond)
        args = {"lid": lid}

        ret = [self._parse_row(row) for row in self._db.execute(sql, args)]
        assert len(ret) > 0, "lid not found"
        assert len(ret) == 1, "lid duplicated"
        return ret[0]

    def update_log(self, d_cond, **kwargs):
        if len(d_cond) == 0:
            # _logger.warning("called update with empty condition")
            raise ValueError("called update with empty condition")

        d_update = self._parse_input(**kwargs)

        table_name = self._valid_table_name(self.tablename_log)
        args = d_cond.copy()
        l_ss = []
        for k, v in d_update.items():
            keyname = "update_" + k
            l_ss.append(db_common.StateSet(k, keyname))
            args[keyname] = v
        l_cond = []
        for c in d_cond.keys():
            if c == "ltgid":
                sql = self._db.select_sql("ltg", ["ltid"],
                                          [db_common.Condition(c, "=", c, True)])
                l_cond.append(db_common.Condition("ltid", "in", sql, False))
            elif c == "dts":
                l_cond.append(db_common.Condition("dt", ">=", c, True))
                args[c] = self._db.strftime(d_cond[c])
            elif c == "dte":
                l_cond.append(db_common.Condition("dt", "<", c, True))
                args[c] = self._db.strftime(d_cond[c])
            else:
                l_cond.append(db_common.Condition(c, "=", c, True))
        sql = self._db.update_sql(table_name, l_ss, l_cond)
        self._db.execute(sql, args)

    def count_lines(self):
        table_name = self.tablename_log
        l_key = ["max(lid)"]
        sql = self._db.select_sql(table_name, l_key)
        cursor = self._db.execute(sql)
        tmp = cursor.fetchone()[0]
        if tmp is None:
            return 0
        else:
            return int(tmp)

    def dt_term(self):
        table_name = self.tablename_log
        l_key = ["min(dt)", "max(dt)"]
        sql = self._db.select_sql(table_name, l_key)
        cursor = self._db.execute(sql)
        top_dtstr, end_dtstr = cursor.fetchone()
        if None in (top_dtstr, end_dtstr):
            raise ValueError("No data found in DB")
        return self._db.datetime(top_dtstr), self._db.datetime(end_dtstr)

    def whole_host_lt(self, dts=None, dte=None):
        table_name = "log"
        l_key = ["host", "ltid"]
        l_cond = []
        args = {}
        if dts is not None:
            l_cond.append(db_common.Condition("dt", ">=", "dts", True))
            args["dts"] = self._db.strftime(dts)
        if dte is not None:
            l_cond.append(db_common.Condition("dt", "<", "dte", True))
            args["dte"] = self._db.strftime(dte)

        sql = self._db.select_sql(table_name, l_key, l_cond, opt=["distinct"])
        cursor = self._db.execute(sql, args)
        return [(row[0], row[1]) for row in cursor]

    def whole_host(self, dts=None, dte=None):
        table_name = "log"
        l_key = ["host"]
        l_cond = []
        args = {}
        if dts is not None:
            l_cond.append(db_common.Condition("dt", ">=", "dts", True))
            args["dts"] = self._db.strftime(dts)
        if dte is not None:
            l_cond.append(db_common.Condition("dt", "<", "dte", True))
            args["dte"] = self._db.strftime(dte)
        sql = self._db.select_sql(table_name, l_key, l_cond, opt=["distinct"])
        cursor = self._db.execute(sql, args)
        return [row[0] for row in cursor]

    def add_lt(self, ltline):
        table_name = self._valid_table_name(self.tablename_lt)
        l_ss = [
            db_common.StateSet("ltid", "ltid"),
            db_common.StateSet("ltw", "ltw"),
            db_common.StateSet("lts", "lts"),
            db_common.StateSet("count", "count")
        ]
        if ltline.lts is None:
            lts = None
        else:
            lts = self._splitter.join(ltline.lts)
        args = {
            "ltid": ltline.ltid,
            "ltw": self._splitter.join(ltline.ltw),
            "lts": lts,
            "count": ltline.count,
        }
        sql = self._db.insert_sql(table_name, l_ss)
        self._db.execute(sql, args)

    def add_ltg(self, ltid, ltgid):
        table_name = self._valid_table_name(self.tablename_ltg)
        l_ss = [
            db_common.StateSet("ltid", "ltid"),
            db_common.StateSet("ltgid", "ltgid")
        ]
        args = {"ltid": ltid, "ltgid": ltgid}
        sql = self._db.insert_sql(table_name, l_ss)
        self._db.execute(sql, args)

    def update_lt(self, ltid, ltw, lts, count=None):
        table_name = self._valid_table_name(self.tablename_lt)
        l_ss = []
        args = {}
        if ltw is not None:
            l_ss.append(db_common.StateSet("ltw", "ltw"))
            args["ltw"] = self._splitter.join(ltw)
        if lts is not None:
            l_ss.append(db_common.StateSet("lts", "lts"))
            args["lts"] = self._splitter.join(lts)
        if count is not None:
            l_ss.append(db_common.StateSet("count", "count"))
            args["count"] = count
        l_cond = [db_common.Condition("ltid", "=", "ltid", True)]
        args["ltid"] = ltid

        sql = self._db.update_sql(table_name, l_ss, l_cond)
        self._db.execute(sql, args)

    def remove_lt(self, ltid):
        args = {"ltid": ltid}

        # remove from lt
        table_name = self._valid_table_name(self.tablename_lt)
        l_cond = [db_common.Condition("ltid", "=", "ltid", True)]
        sql = self._db.delete_sql(table_name, l_cond)
        self._db.execute(sql, args)

        # remove from ltg
        table_name = self._valid_table_name(self.tablename_ltg)
        l_cond = [db_common.Condition("ltid", "=", "ltid", True)]
        sql = self._db.delete_sql(table_name, l_cond)
        self._db.execute(sql, args)

    def restore_lttable(self):
        lttable = lt_common.LTTable()

        table_name = self._db.join_sql("left outer",
                                       self.tablename_lt,
                                       self.tablename_ltg,
                                       "ltid", "ltid")
        l_key = (self.tablename_lt + ".ltid",
                 self.tablename_ltg + ".ltgid",
                 self.tablename_lt + ".ltw",
                 self.tablename_lt + ".lts",
                 self.tablename_lt + ".count")
        sql = self._db.select_sql(table_name, l_key)
        cursor = self._db.execute(sql)
        for row in cursor:
            ltid = int(row[0])
            if row[1]:
                ltgid = int(row[1])
            else:
                ltgid = ltid
            ltw = strutil.split_igesc(row[2], self._splitter)
            tmp = row[3]
            if tmp is None:
                lts = None
            else:
                lts = strutil.split_igesc(tmp, self._splitter)
            count = int(row[4])
            lttable.restore_lt(ltid, ltgid, ltw, lts, count)

        return lttable

    def count_lt(self):
        table_name = self.tablename_lt
        l_key = ["count(*)"]
        sql = self._db.select_sql(table_name, l_key)
        cursor = self._db.execute(sql)
        return int(cursor.fetchone()[0])

    def count_ltg(self):
        table_name = self.tablename_ltg
        l_key = ["max(ltgid)"]
        sql = self._db.select_sql(table_name, l_key)
        cursor = self._db.execute(sql)
        return int(cursor.fetchone()[0]) + 1

    def iter_ltg_def(self):
        table_name = self.tablename_ltg
        l_key = ["ltid", "ltgid"]
        sql = self._db.select_sql(table_name, l_key)
        cursor = self._db.execute(sql)
        for row in cursor:
            ltid, ltgid = row
            yield int(ltid), int(ltgid)

    def iter_ltgid(self):
        table_name = self.tablename_ltg
        l_key = ["ltgid"]
        sql = self._db.select_sql(table_name, l_key, opt=["distinct"])
        cursor = self._db.execute(sql)
        for row in cursor:
            ltgid = row[0]
            yield int(ltgid)

    def get_ltg_members(self, ltgid):
        table_name = "ltg"
        l_key = ["ltid"]
        l_cond = [db_common.Condition("ltgid", "=", "ltgid", True)]
        args = {"ltgid": ltgid}
        sql = self._db.select_sql(table_name, l_key, l_cond)
        cursor = self._db.execute(sql, args)
        return [int(row[0]) for row in cursor]

    def add_tags(self, ltid, tags):
        table_name = self._valid_table_name(self.tablename_tag)
        for tag in tags:
            l_ss = [
                db_common.StateSet("ltid", "ltid"),
                db_common.StateSet("tag", "tag")
            ]
            args = {"ltid": ltid, "tag": tag}
            sql = self._db.insert_sql(table_name, l_ss)
            self._db.execute(sql, args)

    def get_tags(self, **kwargs):
        # compatibility
        l_table_name = self._db.get_table_names()
        if self.tablename_tag not in l_table_name:
            return []

        if "ltgid" in kwargs:
            table_name = self._db.join_sql("left outer",
                                           self.tablename_ltg,
                                           self.tablename_tag,
                                           "ltid", "ltid")
            l_key = [self.tablename_tag + ".tag"]
            l_cond = [db_common.Condition(self.tablename_ltg + ".ltgid",
                                          "=", "ltgid", True)]
            args = {"ltgid": kwargs["ltgid"]}
            sql = self._db.select_sql(table_name, l_key, l_cond, opt=["distinct"])
            cursor = self._db.execute(sql, args)
        elif "ltid" in kwargs:
            table_name = self.tablename_tag
            l_key = ["tag"]
            l_cond = [db_common.Condition("ltid", "=", "ltid", True)]
            args = {"ltid": kwargs["ltid"]}
            sql = self._db.select_sql(table_name, l_key, l_cond)
            cursor = self._db.execute(sql, args)
        else:
            raise ValueError("No identifier given")
        return [row[0] for row in cursor]

    def reset_ltg(self):
        sql = self._db.delete_sql("ltg")
        self._db.execute(sql)

    def reset_tag(self):
        # compatibility
        l_table_name = self._db.get_table_names()
        if self.tablename_tag in l_table_name:
            sql = self._db.delete_sql(self.tablename_tag)
            self._db.execute(sql)
        else:
            self._init_table_tag()
            self._init_index_tag()

    def drop_all(self):
        self._db.reset()


class RestoreOriginalData(object):

    def __init__(self, dirname, style="date", method="commit",
                 reset=False):
        self.style = style
        self.method = method
        self.dirname = dirname
        common.mkdir(dirname)
        if reset:
            _logger.info("RestoreOriginalData was requested to reset files")
            common.rm_dirchild(dirname)

        assert self.style in ("date", "")
        assert self.method in ("incremental", "commit")
        if self.method == "commit":
            self.buf = defaultdict(list)

    def add(self, lm):
        if self.style == "date":
            fn = lm.dt.strftime("%Y%m%d")
        else:
            raise NotImplementedError

        linestr = lm.restore_line()
        if self.method == "incremental":
            self.write_line(linestr, fn)
        elif self.method == "commit":
            self.buf[fn].append(linestr)

    def add_str(self, dt, linestr):
        if self.style == "date":
            fn = dt.strftime("%Y%m%d")
        else:
            raise NotImplementedError

        if self.method == "incremental":
            self.write_line(linestr, fn)
        elif self.method == "commit":
            self.buf[fn].append(linestr)

    def commit(self):
        if self.method == "incremental":
            pass
        elif self.method == "commit":
            self.write_all()

    def write_line(self, linestr, fn):
        with open("/".join((self.dirname, fn)),
                  "a", encoding='utf-8') as f:
            f.write(linestr + "\n")

    def write_all(self):
        assert self.method == "commit"
        for fn, l_buf in self.buf.items():
            with open("/".join((self.dirname, fn)),
                      "a", encoding='utf-8') as f:
                f.write("\n".join(l_buf))


def info(conf):
    """Show abstruction of log messages registered in DB.

    Args:
        conf (config.ExtendedConfigParser): A common configuration object.

    """

    ld = LogData(conf)
    print("[DB status]")
    print("Registered log lines : {0}".format(ld.count_lines()))
    print("Term : {0[0]} - {0[1]}".format(ld.dt_term()))
    print("Log templates : {0}".format(ld.count_lt()))
    print("Log template groups : {0}".format(ld.count_ltg()))
    print("Hosts : {0}".format(len(ld.whole_host())))


def info_term(conf, top_dt, end_dt):
    cnt_line = 0
    s_ltid = set()
    s_gid = set()
    s_host = set()

    ld = LogData(conf)
    for line in ld.iter_lines(top_dt=top_dt, end_dt=end_dt):
        cnt_line += 1
        s_ltid.add(line.lt.ltid)
        s_gid.add(line.lt.ltgid)
        s_host.add(line.host)

    print("[DB status] in {0} - {1}".format(top_dt, end_dt))
    print("Registered log lines : {0}".format(cnt_line))
    print("Log templates : {0}".format(len(s_ltid)))
    print("Log template groups : {0}".format(len(s_gid)))
    print("Hosts : {0}".format(len(s_host)))


def show_all_lt(conf, **kwargs):
    buf = []
    ld = LogData(conf)
    if "simple" in kwargs and kwargs["simple"]:
        for ltobj in ld.iter_lt():
            buf.append(str(ltobj))
    else:
        for ltobj in ld.iter_lt():
            buf.append(ld.show_lt_info(ltobj.ltid))
    return "\n".join(buf)


def show_all_ltg(conf):
    buf = []
    ld = LogData(conf)
    for gid in ld.iter_ltgid():
        buf.append(ld.show_ltg_info(gid))
    return "\n".join(buf)


def show_tag(conf, tag=None):
    d_tag = defaultdict(list)
    ld = LogData(conf)
    for ltobj in ld.iter_lt():
        for tmp_tag in ld.get_tags(ltid=ltobj.ltid):
            if tag is None or tmp_tag == tag:
                d_tag[tmp_tag].append(ltobj.ltid)

    buf = []
    for tmp_tag, l_ltid in d_tag.items():
        buf.append("[tag {0} ({1} tpls)]".format(tmp_tag, len(l_ltid)))
        for ltid in l_ltid:
            buf.append(ld.show_lt_info(ltid))
    return "\n".join(buf)


def show_tag_stats(conf):
    d_tag = defaultdict(list)
    ld = LogData(conf)
    for ltobj in ld.iter_lt():
        for tmp_tag in ld.get_tags(ltid=ltobj.ltid):
            d_tag[tmp_tag].append(ltobj.ltid)

    table = [["tag", "templates", "lines"]]
    for tag, l_ltid in d_tag.items():
        n_line = sum(ld.lt(ltid).count for ltid in l_ltid)
        table.append([tag, len(l_ltid), n_line])
    return common.cli_table(table)


def show_lt_import(conf):
    ld = LogData(conf)
    for ltobj in ld.iter_lt():
        print(" ".join(ltobj.ltw))


def show_all_host(conf, dts=None, dte=None):
    ld = LogData(conf)
    for host in ld.whole_host(dts=dts, dte=dte):
        print(host)


def data_from_db(conf, dirname, method, reset):
    rod = RestoreOriginalData(dirname, method=method, reset=reset)
    ld = LogData(conf)
    top_dt, end_dt = ld.whole_term()
    for lm in ld.iter_lines(top_dt=top_dt, end_dt=end_dt):
        rod.add(lm)
    rod.commit()
